"""Windows utilities."""

from __future__ import annotations

from struct import pack
import enum

from typing_extensions import override

__all__ = (
    'DEFAULT_DPI',
    'LF_FULLFACESIZE',
    'MAX_LINE_LENGTH',
    'CharacterSet',
    'ClipPrecision',
    'Family',
    'Field',
    'NameTooLong',
    'OutputPrecision',
    'Pitch',
    'Quality',
    'Weight',
    'make_font_entry',
)


class Weight(enum.IntEnum):
    """
    The weight of the font in the range 0 through 1000.

    For example, 400 is normal and 700 is bold. If this value is zero, a default weight is used.
    These values are provided for convenience.
    """

    FW_BLACK = 900
    """Black (heavy) weight."""
    FW_BOLD = 700
    """Bold weight."""
    FW_DEMIBOLD = 600
    """Demi-bold (semi-bold) weight."""
    FW_DONTCARE = 0
    """Default weight."""
    FW_EXTRABOLD = 800
    """Extra-bold (ultra-bold) weight."""
    FW_EXTRALIGHT = 200
    """Extra-light (ultra-light) weight."""
    FW_HEAVY = 900
    """Heavy (black) weight."""
    FW_LIGHT = 300
    """Light weight."""
    FW_MEDIUM = 500
    """Medium weight."""
    FW_NORMAL = 400
    """Normal weight."""
    FW_REGULAR = 400
    """Regular weight."""
    FW_SEMIBOLD = 600
    """Semi-bold (demi-bold) weight."""
    FW_THIN = 100
    """Thin weight."""
    FW_ULTRABOLD = 800
    """Ultra-bold (extra-bold) weight."""
    FW_ULTRALIGHT = 200
    """Ultra-light (extra-light) weight."""


class ClipPrecision(enum.IntEnum):
    """
    The clipping precision.

    The clipping precision defines how to clip characters that are partially
    outside the clipping region.
    """

    CLIP_CHARACTER_PRECIS = 0x1
    """Not used."""
    CLIP_DEFAULT_PRECIS = 0x0
    """Specifies default clipping behaviour."""
    CLIP_DFA_DISABLE = 0x40
    """
    Turns off font association for the font. Note that this flag is not guaranteed to have any
    effect on any platform after Windows Server 2003.
    """
    CLIP_EMBEDDED = 0x80
    """You must specify this flag to use an embedded read-only font."""
    CLIP_LH_ANGLES = 0x10
    """
    When this value is used, the rotation for all fonts depends on whether the orientation of the
    coordinate system is left-handed or right-handed. If not used, device fonts always rotate
    counter-clockwise, but the rotation of other fonts is dependent on the orientation of the
    coordinate system.
    """
    CLIP_MASK = 0xF
    """Not used."""
    CLIP_STROKE_PRECIS = 0x2C
    """
    Not used by the font mapper, but is returned when raster, vector, or TrueType fonts are
    enumerated. For compatibility, this value is always returned when enumerating fonts.
    """
    CLIP_TT_ALWAYS = 0x20
    """Not used."""


class CharacterSet(enum.IntEnum):
    """
    The character set.

    Fonts with other character sets may exist in the operating system. If an application uses a font
    with an unknown character set, it should not attempt to translate or interpret strings that are
    rendered with that font.

    This parameter is important in the font mapping process. To ensure consistent results when
    creating a font, do not specify ``OEM_CHARSET`` or ``DEFAULT_CHARSET``. If you specify a
    typeface name in the ``lfFaceName`` member, make sure that the ``lfCharSet`` value matches the
    character set of the typeface specified in ``lfFaceName``.
    """

    ANSI_CHARSET = 0
    """ANSI character set."""
    ARABIC_CHARSET = 178
    """Arabic character set."""
    BALTIC_CHARSET = 186
    """Baltic character set."""
    CHINESEBIG5_CHARSET = 136
    """Chinese Big5 character set."""
    DEFAULT_CHARSET = 1
    """
    DEFAULT_CHARSET is set to a value based on the current system locale. For example, when the
    system locale is English (United States), it is set as ``ANSI_CHARSET``.
    """
    EE_CHARSET = 238
    """Eastern European character set."""
    GB2312_CHARSET = 134
    """GB2312 character set."""
    GREEK_CHARSET = 161
    """Greek character set."""
    HANGUL_CHARSET = 129
    """Hangul (Korean) character set."""
    HEBREW_CHARSET = 177
    """Hebrew character set."""
    JOHAB_CHARSET = 130
    """Johab character set."""
    MAC_CHARSET = 77
    """Macintosh character set."""
    OEM_CHARSET = 255
    """Specifies a character set that is operating-system dependent."""
    RUSSIAN_CHARSET = 204
    """Russian character set."""
    SHIFTJIS_CHARSET = 128
    """Shift JIS character set."""
    SYMBOL_CHARSET = 2
    """Symbol character set."""
    THAI_CHARSET = 222
    """Thai character set."""
    TURKISH_CHARSET = 162
    """Turkish character set."""
    VIETNAMESE_CHARSET = 163
    """Vietnamese character set."""


class OutputPrecision(enum.IntEnum):
    """
    The output precision.

    The output precision defines how closely the output must match the requested font's height,
    width, character orientation, escapement, pitch, and font type.
    """

    OUT_CHARACTER_PRECIS = 2
    """Not used."""
    OUT_DEFAULT_PRECIS = 0
    """Specifies the default font mapper behaviour."""
    OUT_DEVICE_PRECIS = 5
    """
    Instructs the font mapper to choose a Device font when the system contains multiple fonts with
    the same name.
    """
    OUT_OUTLINE_PRECIS = 8
    """
    This value instructs the font mapper to choose from TrueType and other outline-based fonts.
    """
    OUT_PS_ONLY_PRECIS = 10
    """
    Instructs the font mapper to choose from only PostScript fonts. If there are no PostScript
    fonts installed in the system, the font mapper returns to default behaviour.
    """
    OUT_RASTER_PRECIS = 6
    """
    Instructs the font mapper to choose a raster font when the system contains multiple fonts with
    the same name.
    """
    OUT_STRING_PRECIS = 1
    """
    This value is not used by the font mapper, but it is returned when raster fonts are enumerated.
    """
    OUT_STROKE_PRECIS = 3
    """
    This value is not used by the font mapper, but it is returned when TrueType, other
    outline-based fonts, and vector fonts are enumerated.
    """
    OUT_TT_ONLY_PRECIS = 7
    """
    Instructs the font mapper to choose from only TrueType fonts. If there are no TrueType fonts
    installed in the system, the font mapper returns to default behaviour.
    """
    OUT_TT_PRECIS = 4
    """
    Instructs the font mapper to choose a TrueType font when the system contains multiple fonts with
    the same name.
    """


class Pitch(enum.IntEnum):
    """
    The pitch and family of the font.

    In the ``iPitchAndFamily`` argument to ``CreateFontW``, these values represent the two low-order
    bits.
    """

    DEFAULT_PITCH = 0x0
    """Default pitch."""
    FIXED_PITCH = 0x01
    """Fixed pitch (monospace)."""
    MONO_FONT = 0x08
    """Mono font."""
    VARIABLE_PITCH = 0x02
    """Variable pitch (proportional)."""


class Family(enum.IntEnum):
    """
    Font Family.

    Font families describe the look of a font in a general way. They are intended for specifying
    fonts when the exact typeface desired is not available.

    In the ``iPitchAndFamily`` argument to ``CreateFontW``, these values represent bits 4 through 7.
    """

    FF_DECORATIVE = 0x50
    """Novelty fonts. Old English is an example."""
    FF_DONTCARE = 0x0
    """Use default font."""
    FF_MODERN = 0x30
    """
    Fonts with constant stroke width (monospace), with or without serifs. Monospace fonts are
    usually modern. Pica, Elite, and CourierNew are examples.
    """
    FF_ROMAN = 0x10
    """Fonts with variable stroke width (proportional) and with serifs. MS Serif is an example."""
    FF_SCRIPT = 0x40
    """Fonts designed to look like handwriting. Script and Cursive are examples."""
    FF_SWISS = 0x20
    """
    Fonts with variable stroke width (proportional) and without serifs. MS Sans Serif is an
    example.
    """


class Quality(enum.IntEnum):
    """
    Quality.

    The output quality defines how carefully the graphics device interface (GDI) must attempt to
    match the logical-font attributes to those of an actual physical font.

    If neither ``ANTIALIASED_QUALITY`` nor ``NONANTIALIASED_QUALITY`` is selected, the font is
    antialiased only if the user chooses smooth screen fonts in Control Panel.
    """

    ANTIALIASED_QUALITY = 4
    """
    Font is always antialiased if the font supports it and the size of the font is not too small or
    too large.
    """
    CLEARTYPE_QUALITY = 5
    """If set, text is rendered (when possible) using ClearType antialiasing method."""
    DEFAULT_QUALITY = 0
    """Appearance of the font does not matter."""
    DRAFT_QUALITY = 1
    """
    Appearance of the font is less important than when ``PROOF_QUALITY`` is used. For GDI raster
    fonts, scaling is enabled, which means that more font sizes are available, but the quality may
    be lower. Bold, italic, underline, and strikeout fonts are synthesized if necessary.
    """
    NONANTIALIASED_QUALITY = 3
    """Font is never antialiased."""
    PROOF_QUALITY = 2
    """
    Character quality of the font is more important than exact matching of the logical-font
    attributes. For GDI raster fonts, scaling is disabled and the font closest in size is chosen.
    Although the chosen font size may not be mapped exactly when ``PROOF_QUALITY`` is used, the
    quality of the font is high and there is no distortion of appearance. Bold, italic, underline,
    and strikeout fonts are synthesized if necessary.
    """


class Field(str, enum.Enum):
    """Font field names in the registry."""
    @override
    def __str__(self) -> str:
        return self.value

    CaptionFont = 'CaptionFont'
    """Font used in window title bars."""
    IconFont = 'IconFont'
    """Font used for icons in Explorer and similar views."""
    MenuFont = 'MenuFont'
    """Font used for menus."""
    MessageFont = 'MessageFont'
    """Font used in message boxes and many other places."""
    SmCaptionFont = 'SmCaptionFont'
    """Font used in small window title bars."""
    StatusFont = 'StatusFont'
    """Font used in status bars."""


LF_FULLFACESIZE = 64
"""
Maximum face name length for ``LOGFONTW``.

:meta hide-value:
"""
DEFAULT_DPI = 96
"""
Default DPI for font size calculation.

:meta hide-value:
"""
MAX_LINE_LENGTH = 78
"""
Maximum line length in a ``.reg`` file hex value.

:meta hide-value:
"""


class NameTooLong(Exception):
    """Raised when a font name is longer than 64 characters."""
    @override
    def __init__(self, name: str) -> None:
        super().__init__(self, f'{name} length exceeds 64 characters.')


def make_font_entry(
    field: Field,
    name: str = '',
    *,
    char_set: CharacterSet = CharacterSet.DEFAULT_CHARSET,
    clip_precision: ClipPrecision = ClipPrecision.CLIP_DEFAULT_PRECIS,
    default_setting: bool = False,
    header: bool = False,
    dpi: int = DEFAULT_DPI,
    escapement: int = 0,
    font_size_pt: int = 9,
    italic: bool = False,
    orientation: int = 0,
    out_precision: OutputPrecision = OutputPrecision.OUT_DEFAULT_PRECIS,
    pitch_and_family: int = Pitch.VARIABLE_PITCH | Family.FF_SWISS,
    quality: Quality = Quality.DEFAULT_QUALITY,
    strike_out: bool = False,
    underline: bool = False,
    weight: Weight = Weight.FW_NORMAL,
    width: int = 0,
) -> str:
    r"""
    Generate a string for a ``.reg`` file to set a font in ``HKEY_CURRENT_USER\Control Panel\Desktop\WindowMetrics``.

    This is used to set the font for the caption, icon, menu, message, small caption, and status
    fonts. The font name must be less than 64 characters.

    See `LOGFONTW (wingdi.h) - Win32 apps <https://learn.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-logfontw>`
    for more information.

    Parameters
    ----------
    field : Field
        The field name to use in the registry.
    name : str
        The name of the font.
    char_set : CharacterSet
        The character set to use.
    clip_precision : ClipPrecision
        The clip precision to use.
    default_setting : bool
        If ``True``, the header will use the ``HKEY_USERS\.Default`` key. Only applies when
        ``header`` is ``True``.
    header : bool
        If ``True``, the header will be included in the output.
    dpi : int
        The DPI to use. Even if the DPI setting is higher, this should usually be left as ``96``.
    escapement : int
        The escapement to use.
    font_size_pt : int
        The font size in points.
    italic : bool
        If ``True``, the font will be italic.
    orientation : int
        The orientation to use.
    out_precision : OutputPrecision
        The output precision to use.
    pitch_and_family : int
        The pitch and family to use.
    quality : Quality
        The quality to use.
    strike_out : bool
        If ``True``, the font will be struck out.
    underline : bool
        If ``True``, the font will be underlined.
    weight : Weight
        The weight of the font.
    width : int
        The width of the font. This is usually left as ``0``.

    Returns
    -------
    str
        A string composed of multiple lines for use in a ``.reg`` file.

    Raises
    ------
    NameTooLong
        If the font name is longer than 64 characters.
    """  # noqa: E501
    if len(name) > LF_FULLFACESIZE:
        raise NameTooLong(name)
    height = -((font_size_pt * dpi) // DEFAULT_DPI)
    packed = pack(
        '=5l8B64B',
        height,
        width,
        escapement,
        orientation,
        weight,
        italic,
        underline,
        strike_out,
        char_set,
        out_precision,
        clip_precision,
        quality,
        pitch_and_family,
        *name[:LF_FULLFACESIZE].encode('utf-16le').ljust(LF_FULLFACESIZE, b'\0'),
    )
    lines: list[str] = []
    line = f'"{field}"=hex:'
    for n in packed:
        line += f'{n:02x},'
        lc = len(lines)
        if (lc == 0 and len(line) == MAX_LINE_LENGTH) or (lc > 0
                                                          and len(line) == (MAX_LINE_LENGTH - 1)):
            line += '\\'
            lines.append(line)
            line = '  '
    lines.append(line.rstrip(','))
    return '\n'.join((
        *((r'HKEY_USERS\.Default\Control Panel\Desktop\WindowMetrics' if default_setting else
           r'HKEY_CURRENT_USER\Control Panel\Desktop\WindowMetrics') if header else '',),
        *lines,
    ))



"""Internet-related functions."""

from __future__ import annotations

from collections.abc import Callable, Iterator, Sequence
from getpass import getuser
from html import escape
from http import HTTPStatus
from itertools import chain
from os import scandir
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypeAlias, TypedDict, cast
import logging
import os
import plistlib
import re
import urllib.parse

from bs4 import BeautifulSoup as Soup, Tag
from typing_extensions import NotRequired
import keyring
import requests

from .chromium import generate_chrome_user_agent
from .string import hexstr2bytes
from .system import IS_LINUX

if TYPE_CHECKING:
    from .typing import FileDescriptorOrPath, StrPath

__all__ = (
    'BookmarksDataset',
    'BookmarksHTMLAnchorAttributes',
    'BookmarksHTMLFolder',
    'BookmarksHTMLFolderAttributes',
    'BookmarksHTMLLink',
    'RecurseBookmarksHTMLCallback',
    'check_bookmarks_html_urls',
    'generate_html_dir_tree',
    'parse_bookmarks_html',
    'recurse_bookmarks_html',
    'upload_to_imgbb',
    'where_from',
)

log = logging.getLogger(__name__)
KEY_ORIGIN_URL = 'user.xdg.origin.url'
KEY_WHERE_FROMS = 'com.apple.metadata:kMDItemWhereFroms'


def _getxattr(file: FileDescriptorOrPath,
              name: str,
              *,
              follow_symlinks: bool = False) -> bytes:  # pragma: no cover
    from os import getxattr  # noqa: PLC0415

    return getxattr(file, name, follow_symlinks=follow_symlinks)


def where_from(file: FileDescriptorOrPath, *, webpage: bool = False) -> str | None:
    """
    Determine where a file came from based on metadata in extended attributes.

    Parameters
    ----------
    file : FileDescriptorOrPath
        File to inspect.
    webpage : bool
        If ``True``, return the webpage URL instead of the direct download URL (macOS only).

    Returns
    -------
    str | None
        The URL the file was downloaded from, or ``None`` if not available.
    """
    index = 1 if webpage else 0
    attr_value = _getxattr(file, KEY_ORIGIN_URL if IS_LINUX else KEY_WHERE_FROMS).decode()
    if not IS_LINUX:
        return cast('Sequence[str]', plistlib.loads(hexstr2bytes(attr_value)))[index]
    return attr_value


def generate_html_dir_tree(start_dir: StrPath,
                           *,
                           follow_symlinks: bool = False,
                           depth: int = 2) -> str:
    """
    Generate an HTML directory listing.

    Parameters
    ----------
    start_dir : StrPath
        Root directory for the listing.
    follow_symlinks : bool
        If ``True``, follow symbolic links.
    depth : int
        Maximum recursion depth.

    Returns
    -------
    str
        Complete HTML document string.
    """
    def recurse_cwd(path: Path, _cur_depth: int = 0) -> Iterator[str]:
        for entry in sorted(
                sorted(scandir(path), key=lambda x: x.name),
                key=lambda x: not x.is_dir(follow_symlinks=follow_symlinks),
        ):
            if entry.is_dir(follow_symlinks=follow_symlinks) and _cur_depth < depth:
                yield ('<li class="dir mui--text-dark mui--text-body2"><details><summary>'
                       f'<code>{escape(entry.name)}/</code></summary><ul>')
                yield from recurse_cwd(Path(entry), _cur_depth=_cur_depth + 1)
                yield '</ul></details></li>'
            else:
                isd = entry.is_dir(follow_symlinks=follow_symlinks)
                class_ = 'file' if not isd else 'dir'
                slash = '' if not isd else '/'
                yield (
                    f'<li class="{class_} mui--text-dark mui--text-body1"><a class="mui--text-dark"'
                    f' href="./{entry.path}"><code>{escape(entry.name)}{slash}</code></a></li>')

    start_dir = Path(start_dir).resolve(strict=True)
    saved = Path.cwd()
    os.chdir(start_dir)
    try:
        title = start_dir.name
        files = ''.join(chain.from_iterable(recurse_cwd(start_dir)))
        return f"""<!DOCTYPE html>
<html lang="en">
<head>
<meta charset="utf-8">
<meta http-equiv="X-UA-Compatible" content="IE=edge">
<meta name="viewport" content="width=device-width, initial-scale=1">
<title>Contents of {title}/</title>
<link rel="stylesheet" href="https://cdn.muicss.com/mui-0.10.3/css/mui.min.css"
crossorigin="anonymous">
<style>
ul {{
    list-style: none;
    padding-inline-start: 1em;
}}
.dir {{
    cursor: pointer;
}}
.mui-appbar {{
    margin-bottom: 2em;
}}
</style>
</head>
<body>
<header class="mui-appbar mui--z1">
<div class="mui-container-fluid">
<h1 class="mui--text-title">Contents of {title}</h1>
</div>
</header>
<div class="mui-container-fluid">
<ul>{files}</ul>
</div>
</body>
</html>"""
    finally:
        os.chdir(saved)


def upload_to_imgbb(
    path: StrPath,
    *,
    api_key: str | None = None,
    keyring_username: str | None = None,
    timeout: float = 5,
) -> requests.Response:
    """
    Upload an image to ImgBB.

    Get an API key at https://api.imgbb.com/ and set it with ``keyring set imgbb keyring_username``.

    Parameters
    ----------
    path : StrPath
        Path to the image file.
    api_key : str | None
        API key. If ``None``, the key is retrieved from the keyring.
    keyring_username : str | None
        Username for the keyring lookup. Defaults to the current user.
    timeout : float
        HTTP timeout in seconds.

    Returns
    -------
    requests.Response
        The response from the ImgBB API.
    """
    r = requests.post(
        'https://api.imgbb.com/1/upload',
        files={'image': Path(path).resolve(strict=True).read_bytes()},
        params={'key': api_key or keyring.get_password('imgbb', keyring_username or getuser())},
        timeout=timeout,
    )
    r.raise_for_status()
    return r


def stripped_strings_fixed(child: Tag) -> str:
    return re.sub(r'\s+', ' ', ' '.join(child.stripped_strings))


class BookmarksHTMLAnchorAttributes(TypedDict):
    """Attributes of a bookmark link."""

    add_date: str
    """Date the bookmark was added."""
    href: str
    """URL of the bookmark."""
    icon: NotRequired[str]
    """Icon data URI."""


class BookmarksHTMLFolderAttributes(TypedDict):
    """Attributes of a bookmark folder."""

    add_date: str
    """Date the folder was added."""
    last_modified: str
    """Date the folder was last modified."""
    personal_toolbar_folder: NotRequired[Literal['true']]
    """Whether this is the personal toolbar folder."""


class BookmarksHTMLLink(TypedDict):
    """A bookmark link."""

    attrs: BookmarksHTMLAnchorAttributes
    """Link attributes."""
    title: str
    """Link title."""
    type: Literal['link']
    """Type discriminator."""


class BookmarksHTMLFolder(TypedDict):
    """A bookmark folder."""

    attrs: BookmarksHTMLFolderAttributes
    """Folder attributes."""
    children: list[BookmarksHTMLLink | BookmarksHTMLFolder]
    """Child items."""
    name: str
    """Folder name."""
    type: Literal['folder']
    """Type discriminator."""


BookmarksDataset: TypeAlias = list[BookmarksHTMLFolder | BookmarksHTMLLink]
"""
Bookmark top structure.

:meta hide-value:
"""

RecurseBookmarksHTMLCallback: TypeAlias = Callable[
    [BookmarksHTMLAnchorAttributes, str, Sequence[tuple[str, BookmarksHTMLFolderAttributes]]], None]
"""
Callback for :py:func:`recurse_bookmarks_html` to call for each link.

:meta hide-value:
"""


def recurse_bookmarks_html(soup: Tag, callback: RecurseBookmarksHTMLCallback) -> None:
    """
    Traverse the ``bookmarks.html`` tree and call the callback for each link.

    Parameters
    ----------
    soup : Tag
        Parsed HTML tag to traverse.
    callback : RecurseBookmarksHTMLCallback
        Function called for each link found.
    """
    for child in soup.children:
        if not isinstance(child, Tag):
            continue
        match child.name:
            case 'dl' | 'dt' | 'html' | 'body':
                recurse_bookmarks_html(child, callback)
            case 'a':
                folder_path: list[tuple[str, BookmarksHTMLFolderAttributes]] = []
                for parent in child.parents:
                    if parent.name == 'dl' and (h3 := parent.find_previous_sibling('h3')):
                        assert isinstance(h3, Tag)
                        folder_path.append((
                            stripped_strings_fixed(h3),
                            cast('BookmarksHTMLFolderAttributes', h3.attrs),
                        ))
                        break
                callback(
                    cast('BookmarksHTMLAnchorAttributes', child.attrs),
                    stripped_strings_fixed(child),
                    folder_path,
                )


def create_parsed_tree_structure(
        folder_path: Sequence[tuple[str, BookmarksHTMLFolderAttributes]],
        data: BookmarksDataset) -> list[BookmarksHTMLFolder | BookmarksHTMLLink]:
    keys = [f[0] for f in folder_path]
    ref = data
    # This breaks for folders that are named the same at the same depth
    for i, key in enumerate(keys):
        try:
            next(x for x in ref if x['type'] == 'folder' and x['name'] == key)
        except StopIteration:  # noqa: PERF203
            new_level: BookmarksHTMLFolder = {
                'attrs': folder_path[i][1],
                'children': [],
                'name': key,
                'type': 'folder',
            }
            ref.append(new_level)
            ref = new_level['children']
    return ref


def parse_bookmarks_html(html_content: str) -> BookmarksDataset:
    """
    Parse a browser's exported ``bookmarks.html``.

    Parameters
    ----------
    html_content : str
        Raw HTML string.

    Returns
    -------
    BookmarksDataset
        Parsed bookmark tree.
    """
    from bs4 import BeautifulSoup as Soup  # noqa: PLC0415

    data: BookmarksDataset = []

    def callback(
        attrs: BookmarksHTMLAnchorAttributes,
        title: str,
        folder_path: Sequence[tuple[str, BookmarksHTMLFolderAttributes]],
    ) -> None:
        ref = create_parsed_tree_structure(folder_path, data)
        ref.append({'type': 'link', 'title': title, 'attrs': attrs})

    recurse_bookmarks_html(Soup(html_content, 'html5lib'), callback)
    return data


def check_bookmarks_html_urls(
        html_content: str) -> tuple[BookmarksDataset, BookmarksDataset, BookmarksDataset]:
    """
    Check a browser's exported bookmarks.html's URLs.

    Checks for URLs that are not valid any more (status ``404``) or have changed (statuses ``301``
    and ``302``).

    Parameters
    ----------
    html_content : str
        Raw HTML string of the bookmarks file.

    Returns
    -------
    tuple[BookmarksDataset, BookmarksDataset, BookmarksDataset]
        A tuple of ``(all_data, changed, not_found)`` bookmark datasets.
    """
    # After html5lib fixes it, the structure is:
    # DL -> many p -> many DT -> H3 (folder), DL then
    #   p or DT -> A or H3 (folder where structure repeats), DL
    data: BookmarksDataset = []
    changed: BookmarksDataset = []
    not_found: BookmarksDataset = []
    session = requests.Session()
    session.headers.update({
        'cache-control': 'no-cache',
        'dnt': '1',
        'pragma': 'no-cache',
        'referer': 'https://www.google.com/',
        'upgrade-insecure-requests': '1',
        'user_agent': generate_chrome_user_agent(),
    })

    def callback(
        attrs: BookmarksHTMLAnchorAttributes,
        title: str,
        folder_path: Sequence[tuple[str, BookmarksHTMLFolderAttributes]],
    ) -> None:
        ref = create_parsed_tree_structure(folder_path, data)
        new_data: BookmarksHTMLLink = {'type': 'link', 'title': title, 'attrs': attrs}
        if 'href' in attrs and re.match(r'^https?://', attrs['href']):
            log.debug('HEAD %s', attrs['href'])
            r = session.head(attrs['href'])
            if r.status_code in {HTTPStatus.MOVED_PERMANENTLY, HTTPStatus.FOUND}:
                new_location = r.headers['location']
                if not re.match(r'^https://', new_location):
                    parsed = urllib.parse.urlparse(attrs['href'])
                    port_str = f':{parsed.port}' if parsed.port else ''
                    new_location = f'{parsed.scheme}://{parsed.netloc}{port_str}{new_location}'
                log.info(
                    '%d: "%s" @ "%s" -> "%s"',
                    r.status_code,
                    ' / '.join([*(f[0] for f in folder_path), title]),
                    attrs['href'],
                    new_location,
                )
                attrs['href'] = new_location
                changed.append(new_data)
            elif r.status_code == HTTPStatus.NOT_FOUND:
                log.error(
                    '%d: "%s" @ "%s"',
                    r.status_code,
                    ' / '.join([*(f[0] for f in folder_path), title]),
                    attrs['href'],
                )
                not_found.append(new_data)
        ref.append(new_data)

    recurse_bookmarks_html(Soup(html_content, 'html5lib'), callback)
    return data, changed, not_found



"""General I/O functions."""

from __future__ import annotations

from binascii import crc32
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING
from zipfile import ZipFile
import contextlib
import io
import logging
import os
import re
import shutil
import subprocess as sp

from typing_extensions import override

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from .typing import StrPath

__all__ = (
    'RARInfo',
    'SFVVerificationError',
    'UnRAR',
    'UnRARError',
    'UnRARExtractionTestFailed',
    'context_os_open',
    'extract_gog',
    'extract_rar_from_zip',
    'make_sfv',
    'unpack_0day',
    'unpack_ebook',
    'verify_sfv',
)

log = logging.getLogger(__name__)


@contextlib.contextmanager
def context_os_open(path: StrPath,
                    flags: int,
                    mode: int = 511,
                    *,
                    dir_fd: int | None = None) -> Iterator[int]:
    """
    Context-managed file descriptor opener.

    Parameters
    ----------
    path : StrPath
        Path to open.
    flags : int
        Flags to use when opening the file.
    mode : int
        Mode to use when opening the file.
    dir_fd : int | None
        Directory file descriptor to use when opening the file. Defaults to ``None``.

    Yields
    ------
    int
        File descriptor.
    """
    f = os.open(path, flags, mode, dir_fd=dir_fd)
    yield f
    os.close(f)


def unpack_0day(path: StrPath, *, remove_diz: bool = True) -> None:
    """
    Unpack RAR files from 0day zip file sets.

    Parameters
    ----------
    path : StrPath
        Path where zip files are located.
    remove_diz : bool
        Remove any files matching ``*.diz`` glob (not case-sensitive). Defaults to ``True``.
    """
    path = Path(path)
    saved = Path.cwd()
    os.chdir(path)
    try:
        for zip_file in path.glob('*.zip'):
            with ZipFile(zip_file) as z:
                z.extractall()
            zip_file.unlink()
        if remove_diz:
            for diz in path.glob('*.[Dd][Ii][Zz]'):
                diz.unlink()
        rars = list(path.glob('*.rar'))
        with Path(re.sub(r'(?:\.part\d+)?\.r(?:[0-9][0-9]|ar)$', '.sfv',
                         rars[0].name.lower())).open('w+', encoding='utf-8') as f:
            f.write(f'; {datetime.now(tz=timezone.utc).astimezone()}\n')
            f.writelines(f'{rar.name} {crc32(rar.read_bytes()):08X}\n' for rar in sorted(
                path.glob('*.part*.rar' if any(
                    re.search(r'\.part[0-9]{,3}\.rar$', str(r), re.IGNORECASE)
                    for r in rars) else '*.[rstuvwxyz][0-9a][0-9r]')))
    finally:
        os.chdir(saved)


def extract_rar_from_zip(zip_file: ZipFile) -> Iterator[str]:
    """
    Extract RAR files from a zip file.

    Parameters
    ----------
    zip_file : ZipFile
        The zip file to extract RAR files from.

    Yields
    ------
    str
        The name of the extracted RAR file.
    """
    for x in (x for x in zip_file.namelist() if re.search(r'\.r(?:ar|\d{2})$', x)):
        zip_file.extract(x)
        yield x


def unpack_ebook(path: StrPath) -> None:
    """
    Unpack a specially packed eBook file.

    Parameters
    ----------
    path : StrPath
        Path where the zip files are located.

    Raises
    ------
    ValueError
        If no RAR, PDF, or ePub files are found, or if multiple are found unexpectedly.
    NotADirectoryError
        If ``path`` is not a directory.
    FileExistsError
        If no zip files are found.
    """
    def unrar_x(rar: StrPath) -> None:
        sp.run(('unrar', 'x', '-y', str(rar)), capture_output=True, check=True)

    if not (path := Path(path)).is_dir():
        raise NotADirectoryError
    saved = Path.cwd()
    os.chdir(path)
    try:
        zip_listing = frozenset(ZipFile(x) for x in path.iterdir() if x.name.endswith('.zip'))
        if len(zip_listing) == 0:
            raise FileExistsError
        extracted = [Path(x) for y in (extract_rar_from_zip(z) for z in zip_listing) for x in y]
        rar = next((x for x in extracted if x.name.endswith('.rar')), None)
        if not rar:
            raise ValueError(0)
        # Only need the .rar
        unrar_x(rar)
        epub_list = [Path(x) for x in path.iterdir() if x.name.lower().endswith('.epub')]
        pdf_list = [Path(x) for x in path.iterdir() if x.name.lower().endswith('.pdf')]
        if not pdf_list and not epub_list:
            raise ValueError(0)
        if pdf_list:
            if len(pdf_list) > 1:
                log.debug('More than one PDF extracted. Not sure what to do.')
                raise ValueError(len(pdf_list))
            with pdf_list[0].open('rb') as f:
                if (sig := f.read(4)) != b'%PDF':
                    log.warning('PDF file extracted but is not a PDF.')
                    raise ValueError(sig)
            pdf_name = pdf_list[0].resolve(strict=True).parent.name
            ext = 'pdf'
        else:
            if len(epub_list) > 1:
                log.warning('More than one ePub extracted. Not sure what to do.')
                raise ValueError(len(epub_list))
            pdf_name = epub_list[0].resolve(strict=True).parent.name
            ext = 'epub'
            pdf_list = epub_list
        pdf_list[0].rename(f'../{pdf_name}.{ext}')
        for zip_file in zip_listing:
            zip_file.close()
        for x in extracted:
            x.unlink()
    finally:
        os.chdir(saved)


GOG_FILESIZE_RE = re.compile(r'filesizes="(\d+?)"')
GOG_OFFSET_RE = re.compile(r'offset=`head -n (\d+?) "\$0"')


def extract_gog(filename: StrPath, output_dir: StrPath) -> None:
    """
    Extract a Linux gog.com archive.

    Parameters
    ----------
    filename : StrPath
        Path to the GOG archive.
    output_dir : StrPath
        Directory to extract the files to.

    Raises
    ------
    ValueError
        If the file is not a valid GOG archive.
    """
    output_dir = Path(output_dir)
    input_path = Path(filename).resolve(strict=True)
    with input_path.open('rb') as game_bin:
        output_dir.mkdir(parents=True)
        # Read the first 10kb so we can determine the script line number
        beginning = game_bin.read(10240).decode(errors='ignore')
        offset_match = GOG_OFFSET_RE.search(beginning)
        if not offset_match:
            raise ValueError
        script_lines = int(offset_match.group(1))
        # Read the number of lines to determine the script size
        game_bin.seek(0, io.SEEK_SET)
        for _ in range(script_lines):
            game_bin.readline()
        script_size = game_bin.tell()
        log.debug('Makeself script size: %d', script_size)
        # Read the script
        game_bin.seek(0, io.SEEK_SET)
        script_bin = game_bin.read(script_size)
        with (output_dir / 'unpacker.sh').open('wb') as script_f:
            script_f.write(script_bin)
        script = script_bin.decode()
        # Filesize is for the MojoSetup archive, not the actual game data
        filesize_match = GOG_FILESIZE_RE.search(script)
        if not filesize_match:
            raise ValueError
        filesize = int(filesize_match.group(1))
        log.debug('MojoSetup archive size: %d', filesize)
        # Extract the setup archive
        game_bin.seek(script_size, io.SEEK_SET)
        with (output_dir / 'mojosetup.tar.gz').open('wb') as setup_f:
            setup_f.write(game_bin.read(filesize))
        # Extract the game data archive
        dataoffset = script_size + filesize
        game_bin.seek(dataoffset, io.SEEK_SET)
        with (output_dir / 'data.zip').open('wb') as datafile:
            shutil.copyfileobj(game_bin, datafile)


class UnRARError(Exception):
    """General ``unrar`` error."""


class UnRARExtractionTestFailed(UnRARError):
    """Raised when testing extraction fails."""


@dataclass
class RARInfo:
    """File within a RAR information."""

    attributes_str: str
    """File attributes string."""
    date: datetime
    """File date."""
    name: str
    """File name."""
    size: int
    """File size."""


class UnRAR:
    """Simple front-end to an ``unrar`` command."""

    LIST_RE = (r'^\s+'
               r'(?P<attributes>[A-Z\.]{7})\s+'
               r'(?P<size>\d+)\s+'
               r'(?P<date>(?P<year>\d{4})-'
               r'(?P<month>\d{2})-'
               r'(?P<day>\d{2})\s+'
               r'(?P<hour>\d{1,2}):'
               r'(?P<minute>\d{2}))\s+'
               r'(?P<filename>.*)')

    def __init__(self, unrar_path: StrPath = 'unrar') -> None:
        self.unrar_path = str(unrar_path)

    @contextlib.contextmanager
    def pipe(self, rar: StrPath, inner_filename: str) -> Iterator[sp.Popen[bytes]]:
        """
        Start of the pipe of the RAR's content.

        Parameters
        ----------
        rar : StrPath
            Path to the RAR archive.
        inner_filename : str
            Filename within the archive to pipe.

        Yields
        ------
        subprocess.Popen[bytes]
            Handle to the ``unrar`` process.
        """
        with sp.Popen(
            (self.unrar_path, 'p', '-y', '-inul', str(rar), inner_filename),
                stdout=sp.PIPE,
                close_fds=True,
        ) as p:
            yield p

    def test_extraction(self, rar: StrPath, inner_filename: str | None = None) -> None:
        """
        Test extraction.

        Parameters
        ----------
        rar : StrPath
            Path to the RAR archive.
        inner_filename : str | None
            Specific file within the archive to test. If ``None``, tests all files.

        Raises
        ------
        UnRARExtractionTestFailed
            If the extraction test fails.
        """
        try:
            sp.run(
                (
                    self.unrar_path,
                    't',
                    '-y',
                    '-inul',
                    rar,
                    *((inner_filename,) if inner_filename else ()),
                ),
                check=True,
            )
        except sp.CalledProcessError as e:
            raise UnRARExtractionTestFailed from e

    def list_files(self, rar: StrPath) -> Iterator[RARInfo]:
        """
        List files.

        Parameters
        ----------
        rar : StrPath
            Path to the RAR archive.

        Yields
        ------
        RARInfo
            Information about a file in the RAR archive.
        """
        for mm in (m for line in sp.run(
            (self.unrar_path, 'l', '-y',
             rar), text=True, check=True, capture_output=True).stdout.splitlines()
                   if (m := re.match(self.LIST_RE, line))):
            yield RARInfo(
                attributes_str=mm['attributes'],
                date=datetime.strptime(mm['date'], '%Y-%m-%d %H:%M').replace(tzinfo=timezone.utc),
                name=mm['filename'],
                size=int(mm['size']),
            )


class SFVVerificationError(Exception):
    """Raised when SFV fails verification."""
    @override
    def __init__(self, filename: StrPath, expected_crc: int, actual_crc: int) -> None:
        super().__init__(f'{filename}: Expected {expected_crc:08X}. Actual: {actual_crc:08X}.')


def verify_sfv(sfv_file: StrPath) -> None:
    """
    Verify an SFV file.

    Parameters
    ----------
    sfv_file : StrPath
        Path to the SFV file.

    Raises
    ------
    SFVVerificationError
        If the CRC32 checksum does not match.
    """
    sfv_file = Path(sfv_file)
    with sfv_file.open(encoding='utf-8') as f:
        for line in (lj for lj in (li.split(';', 1)[0].split('#', 1)[0].strip() for li in f
                                   if li[0] not in {';', '#'})
                     if re.search(r'[a-z0-9]{08}$', lj, re.IGNORECASE)):
            filename, recorded_crc_s = line.rsplit(' ', 1)
            log.debug('Checking "%s".', filename)
            recorded_crc = int(recorded_crc_s, 16)
            if (crc := crc32((sfv_file.parent / filename).read_bytes())) != recorded_crc:
                raise SFVVerificationError(filename, recorded_crc, crc)


def make_sfv(sfv_file: StrPath, files: Iterable[StrPath], *, header: bool = True) -> None:
    """
    Create an SFV file.

    Parameters
    ----------
    sfv_file : StrPath
        Path to the SFV file.
    files : Iterable[StrPath]
        Files to include in the SFV file.
    header : bool
        If ``True``, include a header with the file size and date. Defaults to ``True``.
    """
    file_paths = sorted([Path(file) for file in files])
    with Path(sfv_file).open('w+', encoding='utf-8') as f:
        if header:
            f.write(f'; Generated on {datetime.now(tz=timezone.utc).isoformat(" ")}\n')
            for file in file_paths:
                stat = file.stat()
                dt = datetime.fromtimestamp(stat.st_mtime, tz=timezone.utc).isoformat(' ')
                f.write(f'; {stat.st_size:-10d} {dt} {file.stem}\n')
        for file in file_paths:
            f.write(f'{file.name} {crc32(file.read_bytes()):08X}\n')



"""Constants."""

from __future__ import annotations

__all__ = ('CONTEXT_SETTINGS',)

CONTEXT_SETTINGS = {'help_option_names': ('-h', '--help')}
"""
Shared context settings for all commands.

:meta hide-value:
"""



from __future__ import annotations

from typing import TYPE_CHECKING

from deltona.commands.git import (
    git_checkout_default_branch_main,
    git_open_main,
    git_rebase_default_branch_main,
    merge_dependabot_prs_main,
)

if TYPE_CHECKING:
    from click.testing import CliRunner
    from pytest_mock import MockerFixture


def test_git_checkout_default_branch_success(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_repo = mocker.patch('deltona.commands.git._get_git_repo')
    mocker.patch('deltona.commands.git.keyring.get_password', return_value='dummy_token')
    mock_get_gh_default_branch = mocker.patch('deltona.commands.git.get_github_default_branch')
    mock_get_gh_default_branch.return_value = 'main'
    mock_checkout = mocker.Mock()
    mock_head = mocker.Mock(checkout=mock_checkout)
    mock_head.name = 'main'
    mock_repo.return_value.heads = [mock_head]

    result = runner.invoke(git_checkout_default_branch_main)
    assert result.exit_code == 0
    mock_get_gh_default_branch.assert_called_once()
    mock_checkout.assert_called_once()


def test_git_checkout_default_branch_no_token(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_repo = mocker.patch('deltona.commands.git._get_git_repo')
    mocker.patch('deltona.commands.git.keyring.get_password', return_value=None)
    mock_get_gh_default_branch = mocker.patch('deltona.commands.git.get_github_default_branch')
    mock_checkout = mocker.Mock()
    mock_head = mocker.Mock(checkout=mock_checkout)
    mock_head.name = 'main'
    mock_repo.return_value.heads = [mock_head]

    result = runner.invoke(git_checkout_default_branch_main)
    assert result.exit_code != 0
    mock_get_gh_default_branch.assert_not_called()
    mock_checkout.assert_not_called()


def test_git_rebase_default_branch_success(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_repo = mocker.patch('deltona.commands.git._get_git_repo')
    mocker.patch('deltona.commands.git.keyring.get_password', return_value='dummy_token')
    mock_get_gh_default_branch = mocker.patch('deltona.commands.git.get_github_default_branch')
    mock_get_gh_default_branch.return_value = 'main'

    result = runner.invoke(git_rebase_default_branch_main, ['--remote'])
    assert result.exit_code == 0
    mock_get_gh_default_branch.assert_called_once()
    mock_repo.return_value.git.rebase.assert_called_once_with('origin/main')


def test_git_rebase_default_branch_no_token(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.git.keyring.get_password', return_value=None)
    mock_get_gh_default_branch = mocker.patch('deltona.commands.git.get_github_default_branch')

    result = runner.invoke(git_rebase_default_branch_main)
    assert result.exit_code != 0
    mock_get_gh_default_branch.assert_not_called()


def test_git_open_main(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_open = mocker.patch('deltona.commands.git.webbrowser.open')
    mock_repo = mocker.patch('deltona.commands.git._get_git_repo')
    mock_repo.return_value.remote.return_value.url = 'https://something'

    result = runner.invoke(git_open_main)
    assert result.exit_code == 0
    mock_open.assert_called_once_with('https://something')


def test_git_open_main_convert_ssh(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_open = mocker.patch('deltona.commands.git.webbrowser.open')
    mock_repo = mocker.patch('deltona.commands.git._get_git_repo')
    mock_repo.return_value.remote.return_value.url = 'git@git.whatever.com:something.git'
    result = runner.invoke(git_open_main)
    assert result.exit_code == 0
    mock_open.assert_called_once_with('https://git.whatever.com/something')


def test_merge_dependabot_prs_main(mocker: MockerFixture, runner: CliRunner) -> None:
    mock_merge = mocker.patch('deltona.commands.git.merge_dependabot_pull_requests',
                              side_effect=[RuntimeError, None])
    mock_sleep = mocker.patch('deltona.commands.git.sleep')
    mocker.patch('deltona.commands.git.keyring.get_password', return_value='dummy_token')

    result = runner.invoke(merge_dependabot_prs_main)
    assert result.exit_code == 0
    assert mock_merge.call_count == 2
    assert mock_sleep.call_count == 1


def test_merge_dependabot_prs_main_no_token(mocker: MockerFixture, runner: CliRunner) -> None:
    mocker.patch('deltona.commands.git.keyring.get_password', return_value=None)
    mock_get_gh_default_branch = mocker.patch('deltona.commands.git.get_github_default_branch')

    result = runner.invoke(merge_dependabot_prs_main)
    assert result.exit_code != 0
    mock_get_gh_default_branch.assert_not_called()



# ruff: noqa: FBT001
from __future__ import annotations

from typing import Union
import os

from deltona.typing import assert_not_none, contains_type_path_like_str
import pytest


@pytest.mark.parametrize(
    ('type_hint', 'expected'),
    [
        (Union[os.PathLike[str], str], True),  # noqa: UP007
        (Union[str, int], False),  # noqa: UP007
        (str, False),
        (Union[os.PathLike[str], int, str], True),  # noqa: UP007
    ],
)
def test_contains_type_path_like_str(type_hint: object, expected: bool) -> None:
    assert contains_type_path_like_str(type_hint) is expected


def test_assert_not_none_returns_value() -> None:
    assert assert_not_none(42) == 42
    assert assert_not_none('hello') == 'hello'


def test_assert_not_none_raises_on_none() -> None:
    with pytest.raises(AssertionError):
        assert_not_none(None)



"""See https://www.sphinx-doc.org/en/master/usage/configuration.html."""

from __future__ import annotations

from datetime import datetime, timezone
from operator import itemgetter
from pathlib import Path
import sys

import tomlkit

with (Path(__file__).parent.parent / 'pyproject.toml').open(newline='\n', encoding='utf-8') as f:
    project_ = tomlkit.load(f).unwrap()['project']
    authors_list, name, version = itemgetter('authors', 'name', 'version')(project_)
authors = [f'{d["name"]} <{d["email"]}>' for d in authors_list]
# region Path setup
# If extensions (or modules to document with autodoc) are in another directory, add these
# directories to sys.path here. If the directory is relative to the documentation root, use
# str(Path().parent.parent) to make it absolute, like shown here.
sys.path.insert(0, str(Path(__file__).parent.parent))
# endregion
author = f'{authors_list[0]["name"]} <{authors_list[0]["email"]}>'
copyright = str(datetime.now(timezone.utc).year)  # noqa: A001
project = name
release = f'v{version}'
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'sphinx_datatables',
    'sphinx_immaterial',
    'sphinxcontrib.autodoc_pydantic',
    'sphinxcontrib.jquery',
]
extensions += ['sphinx_click']
datatables_class = 'sphinx-datatable'
datatables_options = {'paging': 0}
datatables_version = '1.13.4'
html_theme = 'sphinx_immaterial'
html_theme_options = {
    'edit_uri': '/tree/master/docs',
    'features': [
        'announce.dismiss',
        'content.action.edit',
        'content.action.view',
        'content.code.copy',
        'content.tabs.link',
        'content.tooltips',
        'navigation.expand',
        'navigation.footer',
        'navigation.sections',
        'navigation.top',
        'search.share',
        'search.suggest',
        'toc.follow',
        'toc.sticky',
    ],
    'globaltoc_collapse': True,
    'icon': {
        'edit': 'material/file-edit-outline',
        'repo': 'fontawesome/brands/github'
    },
    'palette': [
        {
            'media': '(prefers-color-scheme)',
            'toggle': {
                'icon': 'material/brightness-auto',
                'name': 'Switch to light mode'
            },
        },
        {
            'accent': 'light-blue',
            'media': '(prefers-color-scheme: light)',
            'primary': 'teal',
            'scheme': 'default',
            'toggle': {
                'icon': 'material/lightbulb',
                'name': 'Switch to dark mode'
            },
        },
        {
            'accent': 'blue',
            'media': '(prefers-color-scheme: dark)',
            'primary': 'black',
            'scheme': 'slate',
            'toggle': {
                'icon': 'material/lightbulb-outline',
                'name': 'Switch to system preference'
            },
        },
    ],
    'repo_name': 'deltona',
    'repo_url': 'https://github.com/Tatsh/deltona',
    'site_url': 'https://deltona.readthedocs.org',
    'toc_title_is_page_title': True,
}
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'bascom': ('https://bascom.readthedocs.io/en/latest/', None),
    'binaryornot': ('https://binaryornot.readthedocs.io/en/latest/', None),
    'bs4': ('https://www.crummy.com/software/BeautifulSoup/bs4/doc/', None),
    'click': ('https://click.palletsprojects.com/en/latest/', None),
    'gitpython': ('https://gitpython.readthedocs.io/en/stable/', None),
    'keyring': ('https://keyring.readthedocs.io/en/latest/', None),
    'mutagen': ('https://mutagen.readthedocs.io/en/latest/', None),
    'paramiko': ('https://www.paramiko.org/', None),
    'pexpect': ('https://pexpect.readthedocs.io/en/stable/', None),
    'pillow': ('https://pillow.readthedocs.io/en/stable/', None),
    'platformdirs': ('https://platformdirs.readthedocs.io/en/latest/', None),
    'psutil': ('https://psutil.readthedocs.io/en/latest/', None),
    'pydbus': ('https://pydbus.readthedocs.io/en/latest/', None),
    'pygithub': ('https://pygithub.readthedocs.io/en/latest/', None),
    'pygobject': ('https://pygobject.gnome.org/', None),
    'pyperclip': ('https://pyperclip.readthedocs.io/en/latest/', None),
    'requests': ('https://requests.readthedocs.io/en/latest/', None),
    'typing-extensions': ('https://typing-extensions.readthedocs.io/en/latest/', None),
}



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from unittest import mock
import os

from deltona.system import (
    POSITION_RE,
    STATE_RE,
    MultipleKeySlots,
    find_bluetooth_device_info_by_name,
    get_inhibitor,
    get_kconfig_dict,
    get_kwriteconfig_commands,
    inhibit_notifications,
    kill_gamescope,
    kill_wine,
    pan_connect,
    pan_disconnect,
    patch_macos_bundle_info_plist,
    reset_tpm_enrollment,
    slug_rename,
    uninhibit_notifications,
    wait_for_disc,
)
from deltona.typing import CDStatus
from typing_extensions import Never
import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_mock import MockerFixture

INHIBIT_SUCCESS_CODE = 1234


def make_fake_system_bus() -> tuple[Any, mock.Mock]:
    manager = mock.Mock()

    class FakeSystemBus:
        def get(self, *args: Any, **kwargs: Any) -> dict[str, Any]:  # noqa: PLR6301
            return {'org.freedesktop.login1.Manager': manager}

    return FakeSystemBus, manager


def make_fake_session_bus() -> tuple[Any, mock.Mock]:
    notifications = mock.Mock()

    class FakeSessionBus:
        def get(self, *args: Any, **kwargs: Any) -> dict[str, Any]:  # noqa: PLR6301
            return notifications

    return FakeSessionBus, notifications


def test_get_inhibitor_success(monkeypatch: pytest.MonkeyPatch) -> None:
    fsb, manager = make_fake_system_bus()
    manager.Inhibit.return_value = INHIBIT_SUCCESS_CODE
    monkeypatch.setattr('pydbus.SystemBus', fsb)
    result = get_inhibitor('sleep', 'test_user', 'testing', 'block')
    assert result == INHIBIT_SUCCESS_CODE
    manager.Inhibit.assert_called_once_with('sleep', 'test_user', 'testing', 'block')


def test_inhibit_notifications_success(monkeypatch: pytest.MonkeyPatch) -> None:
    fsb, mock_notifications = make_fake_session_bus()
    monkeypatch.setattr('pydbus.SessionBus', fsb)
    mock_notifications.Inhibited = False
    mock_notifications.Inhibit.return_value = 1234
    result = inhibit_notifications('test_app', 'testing')
    assert result is True
    mock_notifications.Inhibit.assert_called_once_with('test_app', 'testing', {})


def test_inhibit_notifications_already_inhibited(monkeypatch: pytest.MonkeyPatch) -> None:
    fsb, mock_notifications = make_fake_session_bus()
    monkeypatch.setattr('pydbus.SessionBus', fsb)
    mock_notifications.Inhibited = True
    result = inhibit_notifications('test_app', 'testing')
    assert result is False


def test_uninhibit_notifications_success(monkeypatch: pytest.MonkeyPatch) -> None:
    fsb, mock_notifications = make_fake_session_bus()
    monkeypatch.setattr('pydbus.SessionBus', fsb)
    mock_notifications.Inhibited = True
    mock_notifications.UnInhibit.return_value = None
    monkeypatch.setattr('deltona.system._key', 1234)
    uninhibit_notifications()
    mock_notifications.UnInhibit.assert_called_once_with(1234)


def test_uninhibit_notifications_not_inhibited(monkeypatch: pytest.MonkeyPatch) -> None:
    fsb, mock_notifications = make_fake_session_bus()
    monkeypatch.setattr('pydbus.SessionBus', fsb)
    mock_notifications.Inhibited = False
    uninhibit_notifications()
    mock_notifications.UnInhibit.assert_not_called()


def test_uninhibit_notifications_session_bus_get_returns_none(mocker: MockerFixture) -> None:
    mock_session_bus = mocker.patch('pydbus.SessionBus')
    mock_session_bus.return_value.get.return_value = None
    with pytest.raises(ConnectionError):
        uninhibit_notifications()


def test_uninhibit_notifications_key_is_none(mocker: MockerFixture) -> None:
    fsb, mock_notifications = make_fake_session_bus()
    mocker.patch('pydbus.SessionBus', fsb)
    mock_notifications.Inhibited = True
    mocker.patch('deltona.system._key', None)
    uninhibit_notifications()
    mock_notifications.UnInhibit.assert_not_called()


def test_wait_for_disc_success(mocker: MockerFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    def mock_ioctl(*args: Any) -> CDStatus:
        return CDStatus.DISC_OK

    monkeypatch.setattr('fcntl.ioctl', mock_ioctl)
    mock_context_os_open = mocker.patch('deltona.system.context_os_open')
    result = wait_for_disc()
    assert result is True
    mock_context_os_open.assert_called_once_with('dev/sr0', os.O_RDONLY | os.O_NONBLOCK)


def test_wait_for_disc_keyboard_interrupt(mocker: MockerFixture,
                                          monkeypatch: pytest.MonkeyPatch) -> None:
    def mock_ioctl(*args: Any) -> Never:
        raise KeyboardInterrupt

    monkeypatch.setattr('fcntl.ioctl', mock_ioctl)
    mock_context_os_open = mocker.patch('deltona.system.context_os_open')
    assert wait_for_disc() is False
    mock_context_os_open.assert_called_once_with('dev/sr0', os.O_RDONLY | os.O_NONBLOCK)


def test_wait_for_disc_waiting_for_ok(mocker: MockerFixture,
                                      monkeypatch: pytest.MonkeyPatch) -> None:
    call_count = 0

    def mock_ioctl(*args: Any) -> CDStatus:
        nonlocal call_count
        if call_count == 0:
            call_count += 1
            return CDStatus.DRIVE_NOT_READY
        return CDStatus.DISC_OK

    monkeypatch.setattr('fcntl.ioctl', mock_ioctl)
    mocker.patch('deltona.system.context_os_open')
    mock_sleep = mocker.patch('deltona.system.sleep')
    result = wait_for_disc()
    assert result
    assert mock_sleep.call_count == 1


def make_fake_bluez_system_bus() -> tuple[Any, mock.Mock]:
    manager = mock.Mock()

    class FakeSystemBus:
        def get(self, *args: Any, **kwargs: Any) -> dict[str, Any]:  # noqa: PLR6301
            return {'org.freedesktop.DBus.ObjectManager': manager}

    return FakeSystemBus, manager


def test_find_bluetooth_device_info_by_name_success(monkeypatch: pytest.MonkeyPatch) -> None:
    fsb, mock_bluez = make_fake_bluez_system_bus()
    mock_bluez.GetManagedObjects.return_value = {
        '/org/bluez/hci0/dev_00_11_22_33_44_55': {
            'org.bluez.Device1': {
                'Name': 'TestDevice'
            }
        }
    }
    monkeypatch.setattr('pydbus.SystemBus', fsb)
    result = find_bluetooth_device_info_by_name('TestDevice')
    assert result == (  # type: ignore[comparison-overlap]
        '/org/bluez/hci0/dev_00_11_22_33_44_55',
        {
            'Name': 'TestDevice'
        },
    )
    mock_bluez.GetManagedObjects.assert_called_once()


def test_find_bluetooth_device_info_by_name_not_found(monkeypatch: pytest.MonkeyPatch) -> None:
    fsb, mock_bluez = make_fake_bluez_system_bus()
    mock_bluez.GetManagedObjects.return_value = {}
    monkeypatch.setattr('pydbus.SystemBus', fsb)
    with pytest.raises(KeyError):
        find_bluetooth_device_info_by_name('TestDevice')


def test_find_bluetooth_device_info_by_name_not_linux(mocker: MockerFixture) -> None:
    mocker.patch('deltona.system.IS_LINUX', False)  # noqa: FBT003
    with pytest.raises(NotImplementedError):
        find_bluetooth_device_info_by_name('TestDevice')


def test_find_bluetooth_device_info_by_name_no_device1(mocker: MockerFixture) -> None:
    mocker.patch('deltona.system.IS_LINUX', True)  # noqa: FBT003
    fsb, mock_bluez = make_fake_bluez_system_bus()
    mock_bluez.GetManagedObjects.return_value = {
        '/org/bluez/hci0/dev_00_11_22_33_44_55': {
            'org.bluez.NotDevice1': {
                'Name': 'TestDevice'
            }
        }
    }
    mocker.patch('pydbus.SystemBus', fsb)
    with pytest.raises(KeyError):
        find_bluetooth_device_info_by_name('TestDevice')
    mock_bluez.GetManagedObjects.assert_called_once()


def test_slug_rename_success(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_slugify = mocker.patch('deltona.system.slugify')
    mock_path.return_value.resolve.return_value = mock_path
    mock_path.parent = mock_path
    mock_slugify.return_value = 'slugified_name'
    result = slug_rename('test_path')
    assert result == mock_path.rename.return_value
    mock_path.return_value.resolve.assert_called_once_with(strict=True)
    mock_slugify.assert_called_once_with(mock_path.name, no_lower=False)
    mock_path.rename.assert_called_once_with(mock_path / 'slugified_name')


def test_patch_macos_bundle_info_plist_success(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_plistlib = mocker.patch('deltona.system.plistlib')
    mock_info_plist = mock_path.return_value.resolve.return_value.__truediv__.return_value.__truediv__.return_value  # noqa: E501
    mock_info_plist.open.return_value.__enter__.return_value = mock.Mock()
    patch_macos_bundle_info_plist('test_bundle', key='value')
    mock_info_plist.open.assert_any_call('rb')
    mock_info_plist.open.assert_any_call('wb')
    mock_plistlib.load.assert_called_once()
    mock_plistlib.dump.assert_called_once()
    mock_info_plist.touch.assert_called_once()


def test_kill_gamescope_success(monkeypatch: pytest.MonkeyPatch) -> None:
    mock_proc_gs = mock.Mock()
    mock_proc_gs.info = {'name': 'gamescope'}
    mock_proc_gsr = mock.Mock()
    mock_proc_gsr.info = {'name': 'gamescopereaper'}
    mock_proc_not_gs = mock.Mock()
    mock_proc_not_gs.info = {'name': 'not_gamescope'}

    def mock_process_iter(*args: Any) -> Any:
        nonlocal mock_proc_gs, mock_proc_gsr, mock_proc_not_gs
        return [mock_proc_gs, mock_proc_gsr, mock_proc_not_gs]

    monkeypatch.setattr('psutil.process_iter', mock_process_iter)
    kill_gamescope()
    mock_proc_gs.kill.assert_called_once()
    mock_proc_gsr.kill.assert_called_once()
    mock_proc_not_gs.kill.assert_not_called()


def test_pan_connect_success(mocker: MockerFixture) -> None:
    mock_system_bus = mocker.patch('pydbus.SystemBus')
    mock_device = mocker.Mock()
    mock_system_bus.return_value.get.return_value = mock_device
    mocker.patch('deltona.system.IS_LINUX', True)  # noqa: FBT003
    pan_connect('00:11:22:33:44:55', hci='hci1')
    mock_system_bus.return_value.get.assert_called_once_with(
        'org.bluez', '/org/bluez/hci1/dev_00_11_22_33_44_55')
    mock_device.Connect.assert_called_once_with('nap')


def test_pan_connect_not_linux(mocker: MockerFixture) -> None:
    mocker.patch('deltona.system.IS_LINUX', False)  # noqa: FBT003
    with pytest.raises(NotImplementedError):
        pan_connect('00:11:22:33:44:55')


def test_pan_disconnect_success(mocker: MockerFixture) -> None:
    mock_system_bus = mocker.patch('pydbus.SystemBus')
    mock_device = mocker.Mock()
    mock_system_bus.return_value.get.return_value = mock_device
    mocker.patch('deltona.system.IS_LINUX', True)  # noqa: FBT003
    pan_disconnect('00:11:22:33:44:55', hci='hci2')
    mock_system_bus.return_value.get.assert_called_once_with(
        'org.bluez', '/org/bluez/hci2/dev_00_11_22_33_44_55')
    mock_device.Disconnect.assert_called_once_with()


def test_pan_disconnect_not_linux(mocker: MockerFixture) -> None:
    mocker.patch('deltona.system.IS_LINUX', False)  # noqa: FBT003
    with pytest.raises(NotImplementedError):
        pan_disconnect('00:11:22:33:44:55')


def test_kill_wine_kills_wine_processes(mocker: MockerFixture) -> None:
    mock_proc_wineserver = mocker.Mock()
    mock_proc_wineserver.info = {'name': 'wineserver'}
    mock_proc_wine_preloader = mocker.Mock()
    mock_proc_wine_preloader.info = {'name': 'wine-preloader'}
    mock_proc_wine64_preloader = mocker.Mock()
    mock_proc_wine64_preloader.info = {'name': 'wine64-preloader'}
    mock_proc_exe = mocker.Mock()
    mock_proc_exe.info = {'name': 'SOME.EXE'}
    mock_proc_other = mocker.Mock()
    mock_proc_other.info = {'name': 'not_wine'}

    def mock_process_iter(*args: Any, **kwargs: Any) -> list[Any]:
        return [
            mock_proc_wineserver,
            mock_proc_wine_preloader,
            mock_proc_wine64_preloader,
            mock_proc_exe,
            mock_proc_other,
        ]

    mocker.patch('psutil.process_iter', mock_process_iter)
    kill_wine()
    mock_proc_wineserver.kill.assert_called_once()
    mock_proc_wine_preloader.kill.assert_called_once()
    mock_proc_wine64_preloader.kill.assert_called_once()
    mock_proc_exe.kill.assert_called_once()
    mock_proc_other.kill.assert_not_called()


def test_kill_wine_handles_no_matching_processes(mocker: MockerFixture) -> None:
    mock_proc_other1 = mocker.Mock()
    mock_proc_other1.info = {'name': 'foo'}
    mock_proc_other2 = mocker.Mock()
    mock_proc_other2.info = {'name': 'bar'}

    def mock_process_iter(*args: Any, **kwargs: Any) -> list[Any]:
        return [mock_proc_other1, mock_proc_other2]

    mocker.patch('psutil.process_iter', mock_process_iter)
    kill_wine()
    mock_proc_other1.kill.assert_not_called()
    mock_proc_other2.kill.assert_not_called()


def test_kill_wine_handles_case_insensitive_exe(mocker: MockerFixture) -> None:
    mock_proc_exe = mocker.Mock()
    mock_proc_exe.info = {'name': 'program.ExE'}

    def mock_process_iter(*args: Any, **kwargs: Any) -> list[Any]:
        return [mock_proc_exe]

    mocker.patch('psutil.process_iter', mock_process_iter)
    kill_wine()
    mock_proc_exe.kill.assert_called_once()


def test_reset_tpm_enrollment_dry_run_logs(mocker: MockerFixture) -> None:
    mock_log = mocker.patch('deltona.system.log')
    mock_sp_run = mocker.patch('deltona.system.sp.run')
    mock_json_loads = mocker.patch('deltona.system.json.loads')
    mock_json_loads.return_value = {'tokens': {'1': {'type': 'systemd-tpm2', 'keyslots': ['0']}}}
    uuid = 'deadbeef'
    reset_tpm_enrollment(uuid, dry_run=True)
    assert mock_log.info.call_count == 3
    assert mock_sp_run.call_count == 1
    mock_sp_run.assert_called_with(
        ('cryptsetup', 'luksDump', '--dump-json-metadata', f'/dev/disk/by-uuid/{uuid}'),
        check=True,
        capture_output=True,
    )


def test_reset_tpm_enrollment_no_tokens(mocker: MockerFixture) -> None:
    mock_log = mocker.patch('deltona.system.log')
    mock_sp_run = mocker.patch('deltona.system.sp.run')
    mock_json_loads = mocker.patch('deltona.system.json.loads')
    mock_json_loads.return_value = {'tokens': {}}
    uuid = 'deadbeef'
    reset_tpm_enrollment(uuid, dry_run=True)
    mock_log.debug.assert_any_call('No tokens found for device %s.', f'/dev/disk/by-uuid/{uuid}')
    mock_sp_run.assert_called_once()


def test_reset_tpm_enrollment_multiple_keyslots(mocker: MockerFixture) -> None:
    mock_sp_run = mocker.patch('deltona.system.sp.run')
    mock_json_loads = mocker.patch('deltona.system.json.loads')
    mock_json_loads.return_value = {
        'tokens': {
            '1': {
                'type': 'systemd-tpm2',
                'keyslots': ['0', '1']
            }
        }
    }
    uuid = 'deadbeef'
    with pytest.raises(MultipleKeySlots):
        reset_tpm_enrollment(uuid, dry_run=True)
    mock_sp_run.assert_called_once()


def test_reset_tpm_enrollment_real_run(mocker: MockerFixture) -> None:
    mock_log = mocker.patch('deltona.system.log')
    mock_sp_run = mocker.patch('deltona.system.sp.run')
    mock_json_loads = mocker.patch('deltona.system.json.loads')
    mock_json_loads.return_value = {'tokens': {'1': {'type': 'systemd-tpm2', 'keyslots': ['0']}}}
    uuid = 'deadbeef'
    reset_tpm_enrollment(uuid, dry_run=False)
    assert mock_sp_run.call_count == 4
    mock_log.info.assert_called_with('Reset TPM enrolment for %s.', uuid)


def test_get_kwriteconfig_commands_basic(mocker: MockerFixture) -> None:
    mock_default_file = mocker.MagicMock()
    mock_path = mocker.patch('deltona.system.Path')
    mock_path.return_value.resolve.return_value = mock_default_file
    mocker.patch('deltona.system.DEFAULT_FILE', mock_default_file)
    mock_home = mock_path.home.return_value
    mock_home.__str__.return_value = '/home/user'
    mock_path.return_value.exists.return_value = False
    mock_path.return_value.__str__.return_value = '/home/user/.config/kdeglobals'
    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    config_instance.sections.return_value = ['General']
    config_instance.__getitem__.return_value.items.return_value = [('Key', 'Value')]
    mocker.patch('deltona.system.is_binary_string', return_value=False)
    mocker.patch('deltona.system.re.match', side_effect=lambda _, __, **_kw: None)
    mocker.patch('deltona.system.re.search', side_effect=lambda _, __: None)
    commands = list(get_kwriteconfig_commands())
    assert commands
    assert commands[0].startswith('kwriteconfig6')
    assert '--group' in commands[0]
    assert '--key' in commands[0]
    assert 'General' in commands[0]
    assert 'Key' in commands[0]
    assert 'Value' in commands[0]


def test_get_kwriteconfig_commands_not_default_file(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_home = mock_path.home
    mock_home.return_value = mock_path
    mock_path.__truediv__.return_value = mock_path
    mock_path.resolve.return_value = mock_path
    mock_path.__str__.return_value = '/home/user/.config/otherfilerc'  # type: ignore[attr-defined]
    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    config_instance.sections.return_value = ['General']
    config_instance.__getitem__.return_value.items.return_value = [('Key', 'Value')]
    mocker.patch('deltona.system.is_binary_string', return_value=False)
    mocker.patch('deltona.system.Path.exists', return_value=False)
    mocker.patch('deltona.system.re.match', side_effect=lambda _, __, **_kw: None)
    mocker.patch('deltona.system.re.search', side_effect=lambda _, __: None)
    commands = list(get_kwriteconfig_commands('/home/user/.config/otherfilerc'))
    assert commands
    assert commands[0].startswith('kwriteconfig6')
    assert '--group' in commands[0]
    assert '--key' in commands[0]
    assert 'General' in commands[0]
    assert 'Key' in commands[0]
    assert 'Value' in commands[0]


def test_get_kwriteconfig_commands_ignores_binary(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_home = mocker.patch('deltona.system.Path.home')
    mock_home.return_value = mock_path
    mock_path.__truediv__.return_value = mock_path
    mock_path.resolve.return_value = mock_path
    mock_path.__str__.return_value = '/home/user/.config/kdeglobals'  # type: ignore[attr-defined]
    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    config_instance.sections.return_value = ['General']
    config_instance.__getitem__.return_value.items.return_value = [('Key', 'Value')]
    mocker.patch('deltona.system.is_binary_string', return_value=True)
    mocker.patch('deltona.system.Path.exists', return_value=False)
    mocker.patch('deltona.system.re.match', side_effect=lambda _, __, **_kw: None)
    mocker.patch('deltona.system.re.search', side_effect=lambda _, __: None)
    commands = list(get_kwriteconfig_commands('/home/user/.config/kdeglobals'))
    assert commands == []


def test_get_kwriteconfig_commands_ignores_ignored_groups(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_home = mocker.patch('deltona.system.Path.home')
    mock_home.return_value = mock_path
    mock_path.__truediv__.return_value = mock_path
    mock_path.resolve.return_value = mock_path
    mock_path.__str__.return_value = '/home/user/.config/kdeglobals'  # type: ignore[attr-defined]

    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    # Use an ignored group
    config_instance.sections.return_value = ['KFileDialog Settings']
    config_instance.__getitem__.return_value.items.return_value = [('Key', 'Value')]

    mocker.patch('deltona.system.is_binary_string', return_value=False)
    mocker.patch('deltona.system.Path.exists', return_value=False)
    mocker.patch('deltona.system.re.match', side_effect=lambda _, __: None)
    mocker.patch('deltona.system.re.search', side_effect=lambda _, __: None)

    commands = list(get_kwriteconfig_commands('/home/user/.config/kdeglobals'))
    assert commands == []


def test_get_kwriteconfig_commands_skips_metrics_and_state(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_home = mocker.patch('deltona.system.Path.home')
    mock_home.return_value = mock_path
    mock_path.__truediv__.return_value = mock_path
    mock_path.resolve.return_value = mock_path
    mock_path.__str__.return_value = '/home/user/.config/kdeglobals'  # type: ignore[attr-defined]
    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    config_instance.sections.return_value = ['General']
    config_instance.__getitem__.return_value.items.return_value = [
        ('Height', '100'),
        ('State', 'AAAA/'),
        ('Special[$e]', 'foo'),
        ('NormalKey', 'bar'),
    ]

    mocker.patch('deltona.system.is_binary_string', return_value=False)
    mocker.patch('deltona.system.Path.exists', return_value=False)

    def fake_search(pattern: str, value: str) -> bool | None:
        if pattern == POSITION_RE and value == 'Height':
            return True
        if pattern == STATE_RE and value == 'AAAA/':
            return True
        return None

    mocker.patch('deltona.system.re.match', side_effect=lambda _, __, **_kw: None)
    mocker.patch('deltona.system.re.search', side_effect=fake_search)

    commands = list(get_kwriteconfig_commands('/home/user/.config/kdeglobals'))
    # Only 'NormalKey' should be yielded
    assert len(commands) == 1
    assert 'NormalKey' in commands[0]


def test_get_kwriteconfig_commands_type_detection(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_home = mocker.patch('deltona.system.Path.home')
    mock_home.return_value = mock_path
    mock_path.__truediv__.return_value = mock_path
    mock_path.resolve.return_value = mock_path
    mock_path.__str__.return_value = '/home/user/.config/kdeglobals'  # type: ignore[attr-defined]
    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    config_instance.sections.return_value = ['General']
    config_instance.__getitem__.return_value.items.return_value = [
        ('BoolKey', 'true'),
        ('IntKey', '42'),
        ('PathKey', '/some/file'),
        ('StringKey', 'hello'),
    ]
    mocker.patch('deltona.system.is_binary_string', return_value=False)

    def fake_exists(path: Path) -> bool:
        return str(path) == '/some/file'

    mocker.patch('deltona.system.Path.exists', side_effect=fake_exists)

    def fake_match(pattern: str, value: str, *, flags: int = 0) -> bool | None:
        if pattern == r'^(?:1|true|false|on|yes)$' and value == 'true':
            return True
        if pattern == r'^-?[0-9]+$' and value == '42':
            return True
        return None

    mocker.patch('deltona.system.re.match', side_effect=fake_match)
    mocker.patch('deltona.system.re.search', return_value=None)
    commands = list(get_kwriteconfig_commands('/home/user/.config/kdeglobals'))
    assert any('--type' in cmd and 'bool' in cmd for cmd in commands)
    assert any('--type' in cmd and 'int' in cmd for cmd in commands)
    assert any('--type' in cmd and 'path' in cmd for cmd in commands)
    assert any('--type' not in cmd for cmd in commands)  # StringKey has no type


def test_get_kwriteconfig_commands_type_detection_path_not_file(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_path.return_value.exists.side_effect = OSError
    mock_home = mock_path.home
    mock_home.return_value = mock_path
    mock_path.__truediv__.return_value = mock_path
    mock_path.resolve.return_value = mock_path
    mock_path.__str__.return_value = '/home/user/.config/kdeglobals'  # type: ignore[attr-defined]
    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    config_instance.sections.return_value = ['General']
    config_instance.__getitem__.return_value.items.return_value = [
        ('BoolKey', 'true'),
        ('IntKey', '42'),
        ('PathKey', '/some/file'),
        ('StringKey', 'hello'),
    ]
    mocker.patch('deltona.system.is_binary_string', return_value=False)

    def fake_match(pattern: str, value: str, *, flags: int = 0) -> bool | None:
        if pattern == r'^(?:1|true|false|on|yes)$' and value == 'true':
            return True
        if pattern == r'^-?[0-9]+$' and value == '42':
            return True
        return None

    mocker.patch('deltona.system.re.match', side_effect=fake_match)
    mocker.patch('deltona.system.re.search', return_value=None)
    commands = list(get_kwriteconfig_commands('/home/user/.config/kdeglobals'))
    assert any('--type' in cmd and 'bool' in cmd for cmd in commands)
    assert any('--type' in cmd and 'int' in cmd for cmd in commands)
    assert any('--type' in cmd and 'path' not in cmd for cmd in commands)
    assert any('--type' not in cmd for cmd in commands)  # StringKey has no type


def test_get_kwriteconfig_commands_returns_from_unicode_decode_error(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_path.home.return_value = mock_path
    mock_path.__truediv__.return_value = mock_path
    mock_path.resolve.return_value = mock_path
    mock_path.__str__.return_value = '/home/user/.config/kdeglobals'  # type: ignore[attr-defined]
    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    config_instance.read.side_effect = UnicodeDecodeError('utf-8', b'\x80', 0, 1,
                                                          'invalid start byte')
    try:
        list(get_kwriteconfig_commands('/home/user/.config/kdeglobals'))
    except UnicodeDecodeError:
        pytest.fail('Unexpected UnicodeDecodeError raised.')


def test_get_kwriteconfig_commands_skips_sections_with_brackets(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.system.Path')
    mock_home = mocker.patch('deltona.system.Path.home')
    mock_home.return_value = mock_path
    mock_path.__truediv__.return_value = mock_path
    mock_path.resolve.return_value = mock_path
    mock_path.__str__.return_value = '/home/user/.config/kdeglobals'  # type: ignore[attr-defined]
    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    # Section with '][' in the name should be skipped
    config_instance.sections.return_value = ['General][Extra']
    config_instance.__getitem__.return_value.items.return_value = [('Key', 'Value')]
    mocker.patch('deltona.system.is_binary_string', return_value=False)
    mocker.patch('deltona.system.Path.exists', return_value=False)
    mocker.patch('deltona.system.re.match', side_effect=lambda _, __: None)
    mocker.patch('deltona.system.re.search', side_effect=lambda _, __: None)
    commands = list(get_kwriteconfig_commands('/home/user/.config/kdeglobals'))
    assert commands == []


def test_get_kconfig_dict(mocker: MockerFixture) -> None:
    mocker.patch('deltona.system.Path')
    mock_config = mocker.patch('deltona.system.configparser.ConfigParser')
    config_instance = mock_config.return_value
    config_instance.sections.return_value = ['General']
    config_instance.__getitem__.return_value.items.return_value = [
        ('BoolKey', 'true'),
        ('IntKey', '42'),
        ('PathKey', '/some/file'),
        ('StringKey', 'hello'),
    ]

    assert get_kconfig_dict() == {
        'General': {
            'BoolKey': True,
            'IntKey': 42,
            'PathKey': '/some/file',
            'StringKey': 'hello'
        }
    }



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from deltona.chromium import (
    fix_chromium_pwa_icon,
    generate_chrome_user_agent,
    get_last_chrome_major_version,
    get_latest_chrome_major_version,
)
import pytest

if TYPE_CHECKING:
    from pathlib import Path
    from unittest.mock import Mock

    from pytest_mock import MockerFixture
    from requests_mock import Mocker


@pytest.fixture
def mock_pil_image_module(mocker: MockerFixture) -> tuple[Mock, Mock]:
    mock_img = mocker.Mock()
    mock_img.size = (128, 128)
    mock_img.resize.return_value = mock_img
    mock_img.save = mocker.Mock()
    mock_image_mod = mocker.Mock()
    mock_image_mod.open.return_value = mock_img
    mock_image_mod.LANCZOS = 'LANCZOS'
    return mock_image_mod, mock_img


@pytest.fixture
def mock_requests_get(mocker: MockerFixture) -> Mock:
    mock_response = mocker.Mock()
    mock_response.content = b'fake-image'
    mocker.patch('requests.get', return_value=mock_response)
    return cast('Mock', mock_response)


@pytest.fixture
def mock_get_pil_image_module(mocker: MockerFixture, mock_pil_image_module: tuple[Mock,
                                                                                  Mock]) -> Mock:
    mock_image_mod, _ = mock_pil_image_module
    mocker.patch('deltona.chromium._get_pil_image_module', return_value=mock_image_mod)
    return mock_image_mod


def test_fix_chromium_pwa_icon_basic(
    tmp_path: Path,
    mock_get_pil_image_module: Mock,
    mock_requests_get: Mock,
    mock_pil_image_module: tuple[Mock, Mock],
) -> None:
    app_id = 'test_app_id'
    icon_src_uri = 'http://example.com/icon.png'
    config_path = tmp_path
    profile = 'Default'
    mock_img = mock_pil_image_module[1]

    fix_chromium_pwa_icon(config_path, app_id, icon_src_uri, profile)

    # Check that requests.get was called
    mock_requests_get.raise_for_status.assert_called_once()
    # Check that PIL.Image.open was called
    mock_get_pil_image_module.open.assert_called_once()
    # Check that save was called for each size
    assert mock_img.save.call_count > 0
    # Check that files would be saved in the correct directory
    for call in mock_img.save.call_args_list:
        file_path = call.args[0]
        assert 'Icons' in str(file_path)


def test_fix_chromium_pwa_icon_masked(
    tmp_path: Path,
    mock_get_pil_image_module: Mock,
    mock_requests_get: Mock,
    mock_pil_image_module: tuple[Mock, Mock],
) -> None:
    app_id = 'test_app_id'
    icon_src_uri = 'http://example.com/icon.png'
    config_path = tmp_path
    profile = 'Default'
    mock_img = mock_pil_image_module[1]

    fix_chromium_pwa_icon(config_path, app_id, icon_src_uri, profile, masked=True)

    # Should save to both Icons and Icons Maskable
    paths = [call.args[0] for call in mock_img.save.call_args_list]
    assert any('Icons Maskable' in str(p) for p in paths)
    assert any('Icons' in str(p) for p in paths)


def test_fix_chromium_pwa_icon_monochrome(
    mocker: MockerFixture,
    mock_get_pil_image_module: Mock,
    mock_requests_get: Mock,
    mock_pil_image_module: tuple[Mock, Mock],
) -> None:
    mock_path = mocker.patch('deltona.chromium.Path').return_value
    app_id = 'test_app_id'
    icon_src_uri = 'http://example.com/icon.png'
    config_path = mocker.MagicMock()
    config_path.__fspath__.return_value = 'some-config'
    profile = 'Default'
    mock_img = mock_pil_image_module[1]

    fix_chromium_pwa_icon(config_path, app_id, icon_src_uri, profile, monochrome=True)

    # Should save to both Icons and Icons Monochrome
    assert any(x.args[0] for x in mock_path.mock_calls if x.args[0] == 'Icons Monochrome')
    assert any(x.args[0] for x in mock_path.mock_calls if x.args[0] == 'Icons')
    assert mock_img.save.call_count == 8


def test_fix_chromium_pwa_icon_not_square(
    tmp_path: Path,
    mock_get_pil_image_module: Mock,
    mock_requests_get: Mock,
    mock_pil_image_module: tuple[Mock, Mock],
) -> None:
    app_id = 'test_app_id'
    icon_src_uri = 'http://example.com/icon.png'
    config_path = tmp_path
    profile = 'Default'
    mock_img = mock_pil_image_module[1]
    mock_img.size = (128, 64)  # Not square
    with pytest.raises(ValueError, match='Icon is not square'):
        fix_chromium_pwa_icon(config_path, app_id, icon_src_uri, profile)


def test_get_last_chrome_major_version_found(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.chromium.Path')
    mock_instance = mock_path.return_value.expanduser.return_value
    mock_instance.__truediv__.return_value.exists.return_value = True
    mock_instance.__truediv__.return_value.read_text.return_value = '123.0.0.0'
    # Clear cache in case of previous calls
    get_last_chrome_major_version.cache_clear()
    result = get_last_chrome_major_version()
    assert result == '123'


def test_get_last_chrome_major_version_not_found(mocker: MockerFixture) -> None:
    mock_path = mocker.patch('deltona.chromium.Path')
    mock_instance = mock_path.return_value.expanduser.return_value
    mock_instance.__truediv__.return_value.exists.return_value = False
    get_last_chrome_major_version.cache_clear()
    result = get_last_chrome_major_version()
    assert not result


def test_get_latest_chrome_major_version_success(mocker: MockerFixture) -> None:
    # Mock requests.get to return a fake version JSON
    mock_response = mocker.Mock()
    mock_response.json.return_value = {'versions': [{'version': '124.0.6367.60'}]}
    mocker.patch('requests.get', return_value=mock_response)
    get_latest_chrome_major_version.cache_clear()
    result = get_latest_chrome_major_version()
    assert result == '124'


def test_get_latest_chrome_major_version_requests_mock(requests_mock: Mocker) -> None:
    # Use requests-mock to mock the API endpoint
    url = 'https://versionhistory.googleapis.com/v1/chrome/platforms/win/channels/stable/versions'
    requests_mock.get(url, json={'versions': [{'version': '125.0.6422.60'}]})
    get_latest_chrome_major_version.cache_clear()
    result = get_latest_chrome_major_version()
    assert result == '125'


def test_get_latest_chrome_major_version_network_error(mocker: MockerFixture) -> None:
    # Simulate a network error
    mocker.patch('requests.get', side_effect=Exception('Network error'))
    get_latest_chrome_major_version.cache_clear()
    with pytest.raises(Exception, match='Network error'):
        get_latest_chrome_major_version()


def test_generate_chrome_user_agent_with_last_major(mocker: MockerFixture) -> None:
    mocker.patch('deltona.chromium.get_last_chrome_major_version', return_value='123')
    mocker.patch('deltona.chromium.get_latest_chrome_major_version', return_value='999')
    generate_chrome_user_agent.cache_clear()
    ua = generate_chrome_user_agent()
    assert 'Chrome/123.0.0.0' in ua
    assert ua.startswith('Mozilla/5.0 (Windows NT 10.0; Win64; x64)')


def test_generate_chrome_user_agent_with_latest_major(mocker: MockerFixture) -> None:
    mocker.patch('deltona.chromium.get_last_chrome_major_version', return_value='')
    mocker.patch('deltona.chromium.get_latest_chrome_major_version', return_value='456')
    generate_chrome_user_agent.cache_clear()
    ua = generate_chrome_user_agent()
    assert 'Chrome/456.0.0.0' in ua


def test_generate_chrome_user_agent_custom_os(mocker: MockerFixture) -> None:
    mocker.patch('deltona.chromium.get_last_chrome_major_version', return_value='789')
    generate_chrome_user_agent.cache_clear()
    ua = generate_chrome_user_agent('Linux x86_64')
    assert ua.startswith('Mozilla/5.0 (Linux x86_64)')
    assert 'Chrome/789.0.0.0' in ua


def test_generate_chrome_user_agent_cache(mocker: MockerFixture) -> None:
    get_last = mocker.patch('deltona.chromium.get_last_chrome_major_version', return_value='321')
    generate_chrome_user_agent.cache_clear()
    ua1 = generate_chrome_user_agent()
    ua2 = generate_chrome_user_agent()
    assert ua1 == ua2
    get_last.assert_called_once()



"""Configuration for Pytest."""

from __future__ import annotations

from typing import NoReturn
import os

from click.testing import CliRunner
import pytest

if os.getenv('_PYTEST_RAISE', '0') != '0':  # pragma no cover

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call: pytest.CallInfo[None]) -> NoReturn:
        assert call.excinfo is not None
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo: pytest.ExceptionInfo[BaseException]) -> NoReturn:
        raise excinfo.value


@pytest.fixture
def runner() -> CliRunner:
    return CliRunner()



"""Git and Github-related utilities."""

from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse
import contextlib
import logging
import re

if TYPE_CHECKING:
    from git import Repo
    from github.PullRequest import PullRequest
    from github.Repository import Repository

__all__ = (
    'convert_git_ssh_url_to_https',
    'get_github_default_branch',
    'merge_dependabot_pull_requests',
)

log = logging.getLogger(__name__)


def convert_git_ssh_url_to_https(url: str) -> str:
    """
    Convert a Git SSH URI to HTTPS.

    Parameters
    ----------
    url : str
        Git SSH URL to convert.

    Returns
    -------
    str
        The HTTPS equivalent URL.
    """
    if url.startswith('https://'):
        return re.sub(r'\.git$', '', url)
    return re.sub(
        r'\.git$',
        '',
        re.sub(
            r'\.([a-z]+):',
            r'.\1/',
            re.sub(r'^(?:ssh://)?(?:[a-z0-9A-Z]+@)?', 'https://', url, count=1),
            count=1,
        ),
    )


def get_github_default_branch(*,
                              repo: Repo,
                              token: str,
                              base_url: str | None = None,
                              origin_name: str = 'origin') -> str:
    """
    Get the default branch of a GitHub repository.

    Parameters
    ----------
    repo : Repo
        The Git repository.
    token : str
        The GitHub token.
    base_url : str | None
        The base URL of the GitHub API (for enterprise).
    origin_name : str
        The name of the remote to use. Default is 'origin'.

    Returns
    -------
    str
        The default branch of the repository.
    """
    import github  # noqa: PLC0415

    return (github.Github(token, base_url=base_url or github.Consts.DEFAULT_BASE_URL).get_repo(
        urlparse(convert_git_ssh_url_to_https(
            repo.remote(origin_name).url)).path[1:]).default_branch)


def merge_dependabot_pull_requests(*,
                                   token: str,
                                   affiliation: str = 'owner',
                                   base_url: str | None = None) -> None:
    """
    Merge pull requests made by Dependabot on GitHub.

    Parameters
    ----------
    token : str
        The GitHub token.
    affiliation : str
        The affiliation to use. Default is 'owner'.
    base_url : str | None
        The base URL of the GitHub API (for enterprise).

    Raises
    ------
    RuntimeError
        If any pull request could not be merged.
    """
    import github  # noqa: PLC0415

    def uses_dependabot(repo: Repository) -> bool:
        with contextlib.suppress(AttributeError, github.GithubException):
            if repo.security_and_analysis.dependabot_security_updates.status == 'enabled':
                return True
        try:
            repo.get_contents('.github/workflows/dependabot.yml')
        except github.GithubException:
            return False
        else:
            return True

    def has_dependabot_recreate_comment(pull: PullRequest) -> bool:
        try:
            return '@dependabot recreate' in pull.get_issue_comments()[-1].body
        except IndexError:
            return False

    should_raise = False
    for repo in (
            x for x in github.Github(
                token, base_url=base_url or github.Consts.DEFAULT_BASE_URL, per_page=100).get_user(
                ).get_repos(affiliation=affiliation, sort='full_name')  # type: ignore[call-arg]
            if not x.archived and uses_dependabot(x)):
        log.info('Repository: %s', repo.name)
        pull: PullRequest | None = None
        for num in (x.number for x in repo.get_pulls() if x.user.login == 'dependabot[bot]'):
            try:
                pull = repo.get_pull(num)
                if not pull.merge(merge_method='rebase').merged:
                    log.debug('merge() did not raise but merged is False.')
                    if not has_dependabot_recreate_comment(pull):
                        pull.as_issue().create_comment('@dependabot recreate')
            except github.GithubException:
                log.exception('Failed to merge PR %s in repository %s.', num, repo.name)
                if pull and not has_dependabot_recreate_comment(pull):
                    pull.as_issue().create_comment('@dependabot recreate')
                should_raise = True
            pull = None
    if should_raise:
        raise RuntimeError



from __future__ import annotations

from deltona import windows
import pytest


def test_make_font_entry_basic() -> None:
    result = windows.make_font_entry(windows.Field.CaptionFont, name='Arial')
    assert result.startswith('\n"CaptionFont"=hex:')
    assert '41,00,72,00,69,00,61,00,6c,00' in result  # "Arial" in utf-16le hex


def test_make_font_entry_with_all_options() -> None:
    result = windows.make_font_entry(
        field=windows.Field.MenuFont,
        name='Segoe UI',
        char_set=windows.CharacterSet.ANSI_CHARSET,
        clip_precision=windows.ClipPrecision.CLIP_EMBEDDED,
        default_setting=True,
        header=True,
        dpi=96,
        escapement=10,
        font_size_pt=12,
        italic=True,
        orientation=5,
        out_precision=windows.OutputPrecision.OUT_TT_PRECIS,
        pitch_and_family=windows.Pitch.FIXED_PITCH | windows.Family.FF_ROMAN,
        quality=windows.Quality.ANTIALIASED_QUALITY,
        strike_out=True,
        underline=True,
        weight=windows.Weight.FW_BOLD,
        width=2,
    )
    assert result.startswith(r'HKEY_USERS\.Default\Control Panel\Desktop\WindowMetrics')
    assert '"MenuFont"=hex:' in result
    assert '53,00,65,00,67,00,6f,00,65,00,20,00,55,00,49,00' in result  # "Segoe UI" in utf-16le hex


def test_make_font_entry_name_too_long() -> None:
    long_name = 'A' * (windows.LF_FULLFACESIZE + 1)
    with pytest.raises(windows.NameTooLong):
        windows.make_font_entry(windows.Field.MessageFont, name=long_name)


def test_make_font_entry_with_header_current_user() -> None:
    result = windows.make_font_entry(field=windows.Field.StatusFont,
                                     name='Tahoma',
                                     header=True,
                                     default_setting=False)
    assert result.startswith(r'HKEY_CURRENT_USER\Control Panel\Desktop\WindowMetrics')
    assert '"StatusFont"=hex:' in result


def test_make_font_entry_custom_pitch_and_family() -> None:
    pf = windows.Pitch.MONO_FONT | windows.Family.FF_MODERN
    result = windows.make_font_entry(windows.Field.MenuFont, name='Consolas', pitch_and_family=pf)
    assert '43,00,6f,00,6e,00,73,00,6f,00,6c,00,61,00,73,00' in result  # "Consolas" in utf-16le hex



"""Typing helpers."""

from __future__ import annotations

from enum import IntEnum
from os import PathLike
from typing import TYPE_CHECKING, Annotated, Literal, TypeAlias, TypeVar, TypedDict
import os
import typing

from typing_extensions import NotRequired

if TYPE_CHECKING:
    from collections.abc import Sequence

__all__ = (
    'CDStatus',
    'DecodeErrorsOption',
    'FileDescriptorOrPath',
    'INCITS38Code',
    'ProbeDict',
    'StrOrBytesPath',
    'StrPath',
    'StreamDispositionDict',
    'StreamsDict',
    'UNIXStrPath',
    'assert_not_none',
    'contains_type_path_like_str',
)

DecodeErrorsOption: TypeAlias = Literal['ignore', 'replace', 'strict']
"""
Decode errors option for string decoding functions.

:meta hide-value:
"""
INCITS38Code: TypeAlias = Literal[
    'AK',
    'AL',
    'AR',
    'AS',
    'AZ',
    'CA',
    'CO',
    'CT',
    'DC',
    'DE',
    'FL',
    'FM',
    'GA',
    'GU',
    'HI',
    'IA',
    'ID',
    'IL',
    'IN',
    'KS',
    'KY',
    'LA',
    'MA',
    'MD',
    'ME',
    'MH',
    'MI',
    'MN',
    'MO',
    'MP',
    'MS',
    'MT',
    'NC',
    'ND',
    'NE',
    'NH',
    'NJ',
    'NM',
    'NV',
    'NY',
    'OH',
    'OK',
    'OR',
    'PA',
    'PR',
    'PW',
    'RI',
    'SC',
    'SD',
    'TN',
    'TX',
    'UM',
    'UT',
    'VA',
    'VI',
    'VT',
    'WA',
    'WI',
    'WV',
    'WY',
]
"""
Two-letter state code according to INCITS 38-2009.

:meta hide-value:
"""
StrOrBytesPath: TypeAlias = str | bytes | PathLike[str] | PathLike[bytes]
"""
String, bytes, ``PathLike[str]`` or ``PathLike[bytes]``.

:meta hide-value:
"""
StrPath: TypeAlias = str | PathLike[str]
"""
String or ``PathLike[str]``.

:meta hide-value:
"""
FileDescriptorOrPath: TypeAlias = int | StrOrBytesPath
"""
File descriptor or path.

:meta hide-value:
"""
UNIXStrPath: TypeAlias = Annotated[StrPath, 'unix']
"""
String or ``PathLike[str]`` that is a UNIX path.

:meta hide-value:
"""
StrPathMustExist: TypeAlias = Annotated[StrPath, 'must_exist']


class CDStatus(IntEnum):
    """CD status codes."""

    DISC_OK = 4
    """Disc is OK."""
    DRIVE_NOT_READY = 3
    """Drive is not ready."""
    NO_DISC = 1
    """No disc in drive."""
    NO_INFO = 0
    """No information available."""
    TRAY_OPEN = 2
    """Tray is open."""


def contains_type_path_like_str(type_hints: object) -> bool:
    """
    Check if a type hint contains ``os.PathLike[str]``.

    Parameters
    ----------
    type_hints : object
        The type hint to inspect.

    Returns
    -------
    bool
        ``True`` if the type hint contains ``os.PathLike[str]``.
    """
    return os.PathLike[str] in typing.get_args(type_hints)


_T = TypeVar('_T')


def assert_not_none(var: _T | None) -> _T:
    """
    Assert the ``var`` is not None and return it.

    This will remove ``None`` from type ``_T | None``.

    Parameters
    ----------
    var : T | None
        The variable to check.

    Returns
    -------
    T
        The variable, guaranteed to be not None.
    """
    assert var is not None
    return var


# Used by chrome-bisect-flags
class ChromeLocalStateBrowser(TypedDict):
    enabled_labs_experiments: Sequence[str]


class ChromeLocalState(TypedDict):
    browser: ChromeLocalStateBrowser


class StreamDispositionDict(TypedDict):
    """FFmpeg stream disposition dictionary."""

    default: Literal[0, 1]
    """Default stream."""


class TagsDict(TypedDict):
    info_json: NotRequired[str]
    TXXX: NotRequired[str]


class StreamsDict(TypedDict):
    """FFmpeg stream dictionary."""

    codec_type: Literal['audio', 'video']
    """Codec type."""
    disposition: StreamDispositionDict
    """Stream disposition dictionary."""
    height: int
    """Height of the video stream."""
    tags: TagsDict
    """Tags dictionary."""
    width: int
    """Width of the video stream."""


class FormatDict(TypedDict):
    tags: TagsDict


class ProbeDict(TypedDict):
    """FFmpeg probe result returned by :py:func:`deltona.media.ffprobe`."""

    format: FormatDict
    """Format dictionary."""
    streams: Sequence[StreamsDict]
    """Dictionary of streams."""



from __future__ import annotations

from typing import TYPE_CHECKING
import json

from deltona.commands.www import (
    check_bookmarks_html_main,
    chrome_bisect_flags_main,
    fix_chromium_pwa_icon_main,
    where_from_main,
)

if TYPE_CHECKING:
    from pathlib import Path

    from click.testing import CliRunner
    from pytest_mock import MockerFixture


def test_where_from_single(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mock_where_from = mocker.patch('deltona.commands.www.where_from')
    mock_where_from.return_value = 'where_from_result'
    mock_arg1 = tmp_path / 'arg1.txt'
    mock_arg1.write_text('This is a test file.')
    result = runner.invoke(where_from_main, [str(mock_arg1)])
    assert result.exit_code == 0
    assert 'where_from_result' in result.output


def test_where_from_multiple(mocker: MockerFixture, runner: CliRunner, tmp_path: Path) -> None:
    mock_where_from = mocker.patch('deltona.commands.www.where_from')
    mock_where_from.return_value = 'where_from_result'
    mock_arg1 = tmp_path / 'arg1.txt'
    mock_arg1.write_text('This is a test file.')
    mock_arg2 = tmp_path / 'arg2.txt'
    mock_arg2.write_text('This is another test file.')
    result = runner.invoke(where_from_main, [str(mock_arg1), str(mock_arg2)])
    assert result.exit_code == 0
    assert 'arg1.txt: where_from_result' in result.output
    assert 'arg2.txt: where_from_result' in result.output


def test_fix_chromium_pwa_icon_main(mocker: MockerFixture, runner: CliRunner,
                                    tmp_path: Path) -> None:
    mock_fix_icon = mocker.patch('deltona.commands.www.fix_chromium_pwa_icon')
    result = runner.invoke(fix_chromium_pwa_icon_main, ['app_id', 'src_uri', '-c', str(tmp_path)])
    assert result.exit_code == 0
    mock_fix_icon.assert_called_once_with(mocker.ANY,
                                          'app_id',
                                          'src_uri',
                                          'Default',
                                          masked=False,
                                          monochrome=False)


def test_check_bookmarks_html_main(mocker: MockerFixture, runner: CliRunner,
                                   tmp_path: Path) -> None:
    mock_check_urls = mocker.patch(
        'deltona.commands.www.check_bookmarks_html_urls',
        return_value=[None, ['a', 'b'], ['c', 'd']],
    )
    mock_file = tmp_path / 'bookmarks.html'
    mock_file.write_text('<html><body>Bookmarks</body></html>')
    result = runner.invoke(check_bookmarks_html_main, [str(mock_file)])
    assert result.exit_code == 0
    mock_check_urls.assert_called_once_with('<html><body>Bookmarks</body></html>')
    assert '2 URLs changed.' in result.output
    assert '2 URLs resulted in 404 response.' in result.output


def test_chrome_bisect_flags_main_no_flags(mocker: MockerFixture, runner: CliRunner,
                                           tmp_path: Path) -> None:
    mock_local_state = tmp_path / 'Local State'
    mock_local_state.write_text(json.dumps({'browser': {'enabled_labs_experiments': []}}))
    result = runner.invoke(chrome_bisect_flags_main, [str(mock_local_state)])
    assert result.exit_code != 0
    assert 'Nothing to test.' in result.output


def test_chrome_bisect_flags_main_3_flags(mocker: MockerFixture, runner: CliRunner,
                                          tmp_path: Path) -> None:
    mocker.patch('deltona.commands.www.kill_processes_by_name')
    mock_local_state = tmp_path / 'Local State'
    mock_local_state.write_text(
        json.dumps({'browser': {
            'enabled_labs_experiments': ['a', 'b', 'c']
        }}))
    result = runner.invoke(chrome_bisect_flags_main, [str(mock_local_state)], input='\n\ny\nn\ny\n')
    assert result.exit_code == 0
    assert 'Saved "Local State" with "a" removed.' in result.output


def test_chrome_bisect_flags_main_4_flags(mocker: MockerFixture, runner: CliRunner,
                                          tmp_path: Path) -> None:
    mocker.patch('deltona.commands.www.kill_processes_by_name')
    mock_local_state = tmp_path / 'Local State'
    mock_local_state.write_text(
        json.dumps({'browser': {
            'enabled_labs_experiments': ['a', 'b', 'c', 'd']
        }}))
    result = runner.invoke(chrome_bisect_flags_main, [str(mock_local_state)],
                           input=('\n\nn\n\ny\n\nn\n\ny\n'))
    assert result.exit_code == 0
    assert 'Saved "Local State" with "d" removed.' in result.output


def test_chrome_bisect_flags_main_invalid_response(mocker: MockerFixture, runner: CliRunner,
                                                   tmp_path: Path) -> None:
    mocker.patch('deltona.commands.www.kill_processes_by_name')
    mock_local_state = tmp_path / 'Local State'
    mock_local_state.write_text(
        json.dumps({'browser': {
            'enabled_labs_experiments': ['a', 'b', 'c', 'd']
        }}))
    result = runner.invoke(chrome_bisect_flags_main, [str(mock_local_state)],
                           input=('\n\nn\n\ny\n\nn\n\n\n'))  # Invalid response here.
    assert result.exit_code == 0
    assert 'Restored original "Local State".' in result.output


def test_chrome_bisect_flags_main_single_flag(mocker: MockerFixture, runner: CliRunner,
                                              tmp_path: Path) -> None:
    mocker.patch('deltona.commands.www.kill_processes_by_name')
    mock_local_state = tmp_path / 'Local State'
    mock_local_state.write_text(json.dumps({'browser': {'enabled_labs_experiments': ['a']}}))
    result = runner.invoke(chrome_bisect_flags_main, [str(mock_local_state)], input='\n\nn\n')
    assert result.exit_code == 0
    assert 'Saved "Local State" with "a" removed.' in result.output



from __future__ import annotations

from typing import TYPE_CHECKING

from deltona.commands.main import main

if TYPE_CHECKING:
    from click.testing import CliRunner
    from pytest_mock import MockerFixture

_LINUX_ONLY_COMMANDS = (
    'clean-old-kernels-modules',
    'connect-g603',
    'inhibit-notifications',
    'kill-gamescope',
    'systemd-reset-tpm-cryptenroll',
    'wait-for-disc',
)
_WINE_COMMANDS = (
    'kill-wine',
    'set-wine-fonts',
    'unregister-wine-assocs',
    'winegoginstall',
    'wineshell',
)


def test_main_help(runner: CliRunner) -> None:
    result = runner.invoke(main, ['--help'])
    assert result.exit_code == 0
    assert 'Deltona' in result.output


def test_main_list_commands(runner: CliRunner) -> None:
    result = runner.invoke(main, ['--help'])
    assert result.exit_code == 0
    for cmd_name in ('add-cdda-times', 'adp', 'slugify', 'urldecode'):
        assert cmd_name in result.output


def test_main_subcommand_help(runner: CliRunner) -> None:
    result = runner.invoke(main, ['add-cdda-times', '--help'])
    assert result.exit_code == 0
    assert 'CDDA' in result.output


def test_main_unknown_subcommand(runner: CliRunner) -> None:
    result = runner.invoke(main, ['no-such-command'])
    assert result.exit_code != 0


def test_main_subcommand_execution(runner: CliRunner) -> None:
    result = runner.invoke(main, ['add-cdda-times', '01:02:73', '02:05:09'])
    assert result.exit_code == 0
    assert '03:08:07' in result.output


def test_main_shows_all_commands_on_linux(runner: CliRunner, mocker: MockerFixture) -> None:
    mocker.patch('deltona.commands.main.sys.platform', 'linux')
    result = runner.invoke(main, ['--help'])
    assert result.exit_code == 0
    for cmd in (*_LINUX_ONLY_COMMANDS, *_WINE_COMMANDS):
        assert cmd in result.output


def test_main_hides_linux_only_on_darwin(runner: CliRunner, mocker: MockerFixture) -> None:
    mocker.patch('deltona.commands.main.sys.platform', 'darwin')
    result = runner.invoke(main, ['--help'])
    assert result.exit_code == 0
    for cmd in _LINUX_ONLY_COMMANDS:
        assert cmd not in result.output
    for cmd in _WINE_COMMANDS:
        assert cmd in result.output
    assert 'add-cdda-times' in result.output


def test_main_hides_not_windows_on_win32(runner: CliRunner, mocker: MockerFixture) -> None:
    mocker.patch('deltona.commands.main.sys.platform', 'win32')
    result = runner.invoke(main, ['--help'])
    assert result.exit_code == 0
    for cmd in (*_LINUX_ONLY_COMMANDS, *_WINE_COMMANDS):
        assert cmd not in result.output
    assert 'add-cdda-times' in result.output



"""Name utilities."""

from __future__ import annotations

from enum import IntEnum
from typing import TYPE_CHECKING
import re

from .string import fix_apostrophes, is_roman_numeral

if TYPE_CHECKING:
    from collections.abc import Iterable

__all__ = ('Mode', 'adjust_title')

# Non-strict, not including words like below, forms of to be, forms of you/he/etc, or words like
# 'call'.
STOP_WORDS = {
    'a',
    'an',
    'and',
    'at',
    'by',
    'de',  # mainly for Spanish and French
    'el',  # Spanish
    'feat',
    'featuring',
    'for',
    'from',
    'il',  # Italian
    'in',
    'into',
    'la',  # Spanish/French/Italian
    'lo',  # Italian
    'of',
    'off',
    'on',
    'or',
    'per',
    'por',  # Spanish
    # 'so',
    'te',  # Spanish/French
    # 'than',
    'the',
    # 'then',
    # 'this',
    'to',
    # 'too',
    'van',
    'via',
    'von',
    'vs',
    'with',
    'within',
    'without',
}
ENGLISH_ORDINAL_RE = r'(\d+)(st|nd|rd|th)'
ENGLISH_ABBREV = {'feat', 'mr', 'mrs', 'ms', 'vs'}
JAPANESE_PARTICLES = {'de', 'e', 'ga', 'ha', 'ka', 'kana', 'ne', 'ni', 'no', 'to', 'wa', 'wo'}
CHINESE_PARTICLES = {'de', 'ge', 'he', 'le', 'ma'}
ARABIC_STOPS = {
    'al',
    'ala',
    'alayhi',
    'alayka',
    'alayya',
    'an',
    'anhu',
    'anka',
    'anni',
    'bi',
    'biha',
    'bihi',
    'bika',
    'fi',
    'fihi',
    'fika',
    'fiya',
    'ila',
    'ilayhi',
    'ilayka',
    'ilayya',
    'lahu',
    'laka',
    'li',
    'maa',
    'maahu',
    'maaka',
    'mai',
    'min',
    'minhu',
    'minka',
    'minni',
    'wa',
}
NAMES_TO_FIX = {
    "mcdonald's": "McDonald's",
    'Arkit': 'ARKit',
    'Imessage': 'iMessage',
    'Ios': 'iOS',  # Not Cisco IOS
    'Itunes': 'iTunes',
    'LLVM': 'LLVM',
    'Macos': 'macOS',
    'Mapkit': 'MapKit',
    'Pdfkit': 'PDFKit',
    'S3RL': 'S3RL',
    'Sirikit': 'SiriKit',
    'Tvos': 'tvOS',
    'Watchos': 'watchOS',
    'Whats': "What's",
    'Wkwebview': 'WKWebView',
    'Wwdc': 'WWDC',
    'mcdonald': 'McDonald',
    'mcdonalds': "McDonald's",
}


class Mode(IntEnum):
    """Mode to operate in."""

    English = 1
    """
    English mode.

    :meta hide-value:
    """
    Japanese = 1 << 1
    """
    Japanese mode.

    :meta hide-value:
    """
    Chinese = 1 << 2
    """
    Chinese mode.

    :meta hide-value:
    """
    Arabic = 1 << 3
    """
    Arabic mode.

    :meta hide-value:
    """


MODE_MAP = {
    Mode.Arabic: ARABIC_STOPS,
    Mode.Chinese: CHINESE_PARTICLES,
    Mode.English: STOP_WORDS,
    Mode.Japanese: JAPANESE_PARTICLES,
}


def _get_name(word: str, names: dict[str, str] = NAMES_TO_FIX) -> str | None:
    word = word.lower()
    for name, output in names.items():
        if name.lower() == word:
            return output
    return None


def adjust_title(
    words: str,
    modes: Iterable[Mode] = (Mode.English,),
    names: dict[str, str] = NAMES_TO_FIX,
    *,
    disable_names: bool = False,
    ampersands: bool = False,
) -> str:
    """
    Adjust a string that represents a title.

    Primarily for English to lowercase stop words, but also works for other languages. Some name
    rules are built-in but can also be passed in the ``names`` parameter.

    It is far from perfect.

    Parameters
    ----------
    words : str
        The string to adjust.
    modes : Iterable[Mode]
        The modes to operate in. Default is English.
    names : dict[str, str]
        A dictionary of names to fix. The key is the name to fix, and the value is the fixed name.
        Default is a set of common names.
    disable_names : bool
        If True, do not fix names. Default is ``False``.
    ampersands : bool
        If True, replace ``' and '`` with ``' & '``. Default is ``False``.

    Returns
    -------
    str
        The adjusted string.
    """
    original_words = words.strip().split()
    word_list = words.strip().title().split()
    name = _get_name(word_list[0], names=names) if not disable_names else None
    if name:
        title = [name]
    elif word_list[0].upper() == original_words[0]:
        title = [fix_apostrophes(original_words[0])]
    else:
        title = [fix_apostrophes(word_list[0])]
    if is_roman_numeral(title[0]):
        title = [title[0].upper()]
    last_index = len(original_words) - 1
    # MIX is a roman numeral but is more typically used in a sequence like 'Extended Mix', so do not
    # capitalise it.
    ignore_roman = {'mix'}
    for mode in modes:
        to_lower_case_array = MODE_MAP[mode]
        index = 1
        for word in word_list[1:]:
            new_word = word
            if not disable_names and (name := _get_name(new_word, names=names)):
                try:
                    title[index] = name
                except IndexError:
                    title.append(name)
                index += 1
                continue
            # Detect an upper-case new_word not to change.
            # First detect I and ignore it, then detect uppercase letters and numbers.
            if (original_words[index] == new_word.upper() and not re.match(r'[^\w]', new_word)
                ) and (not (mode == Mode.English and new_word == 'I') and
                       (index == last_index and re.match(r'[A-Z0-9]+', original_words[index]))):
                title.append(original_words[index])
                continue
            begin = end = ''
            if begin_match := re.match(r'^(\W+)', new_word):
                begin = begin_match.group(0)
                new_word = new_word[len(begin):]
            if end_match := re.search(r'(\W+)$', new_word):
                end = end_match.group(0)
                new_word = new_word[0:-len(end)]
            if new_word.lower() in to_lower_case_array:
                new_word = word_list[index] = new_word.lower()
            new_word = fix_apostrophes(new_word)
            if is_roman_numeral(new_word) and new_word.lower() not in ignore_roman:
                new_word = new_word.upper()
            if mode == Mode.English and new_word.lower() in ENGLISH_ABBREV:
                end = ''
            ordinal_match = re.match(ENGLISH_ORDINAL_RE, new_word, flags=re.IGNORECASE)
            if mode == Mode.English and ordinal_match is not None:
                new_word = ordinal_match.group(1) + ordinal_match.group(2).lower()
            new_word = f'{begin}{new_word}{end}'
            try:
                title[index] = new_word
            except IndexError:
                title.append(new_word)
            index += 1
    title_ = ' '.join(title)
    if ampersands:
        title_ = title_.replace(' and ', ' & ')
    return title_



# ruff: noqa: RUF001
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from deltona import string
import pytest

if TYPE_CHECKING:
    from collections.abc import Sequence

    from pytest_mock import MockerFixture


def test_strip_ansi_removes_ansi_sequences() -> None:
    ansi_str = '\x1b[31mRed Text\x1b[0m'
    assert string.strip_ansi(ansi_str) == 'Red Text'


def test_strip_ansi_returns_original_if_no_ansi() -> None:
    plain_str = 'No ANSI here'
    assert string.strip_ansi(plain_str) == plain_str


def test_strip_ansi_empty_string() -> None:
    assert string.strip_ansi('') == ''  # noqa: PLC1901


def test_underscorize_replaces_spaces_with_underscore() -> None:
    s = 'hello world test'
    result = string.underscorize(s)
    assert result == 'hello_world_test'


def test_underscorize_replaces_tabs_and_newlines() -> None:
    s = 'hello\tworld\nfoo'
    result = string.underscorize(s)
    assert result == 'hello_world_foo'


def test_underscorize_multiple_whitespace() -> None:
    s = 'a   b\t\tc\n\nd'
    result = string.underscorize(s)
    assert result == 'a_b_c_d'


def test_underscorize_empty_string() -> None:
    s = ''
    result = string.underscorize(s)
    assert result == ''  # noqa: PLC1901


def test_underscorize_no_whitespace() -> None:
    s = 'nowhitespace'
    result = string.underscorize(s)
    assert result == 'nowhitespace'


def test_is_ascii_all_ascii() -> None:
    s = 'hello world'
    assert string.is_ascii(s) is True


def test_is_ascii_with_non_ascii() -> None:
    # cspell: disable-next-line  # noqa: ERA001
    s = 'hello wørld'
    assert string.is_ascii(s) is False


def test_is_ascii_empty_string() -> None:
    s = ''
    assert string.is_ascii(s) is True


def test_is_ascii_all_non_ascii() -> None:
    s = 'øæå'
    assert string.is_ascii(s) is False


def test_is_ascii_with_numbers_and_symbols() -> None:
    s = '1234!@#$'
    assert string.is_ascii(s) is True


def test_is_ascii_with_surrogate_pair() -> None:
    # Simulate a string with a high codepoint (emoji)
    s = 'hello \U0001f600'
    assert string.is_ascii(s) is False


def test_hexstr2bytes_valid_hex() -> None:
    s = '01020a'
    result = string.hexstr2bytes(s)
    assert result == bytes([1, 2, 10])


def test_hexstr2bytes_empty_string() -> None:
    s = ''
    result = string.hexstr2bytes(s)
    assert result == b''


def test_hexstr2bytes_invalid_hex_raises() -> None:
    s = '01z'
    with pytest.raises(ValueError, match='z'):
        string.hexstr2bytes(s)


def test_hexstr2bytes_generator_yields_correct_values() -> None:
    s = '0a1b2c'
    gen = string.hexstr2bytes_generator(s)
    assert list(gen) == [10, 27, 44]


def test_unix_path_to_wine_existing_path(tmp_path: Path) -> None:
    # Create a file and get its path
    file_path = tmp_path / 'file.txt'
    file_path.write_text('test')
    result = string.unix_path_to_wine(str(file_path))
    # Should start with Z: and use backslashes
    assert result.startswith('Z:')
    assert '\\' in result
    assert 'file.txt' in result


def test_unix_path_to_wine_non_existent_path(tmp_path: Path) -> None:
    non_existent = tmp_path / 'does_not_exist.txt'
    result = string.unix_path_to_wine(str(non_existent))
    assert result.startswith('Z:')
    assert '\\' in result
    assert 'does_not_exist.txt' in result


def test_unix_path_to_wine_relative_path() -> None:
    # Use a relative path
    rel_path = 'some/relative/path.txt'
    cwd = Path.cwd()
    expected = 'Z:' + str(cwd).replace('/', '\\') + '\\some\\relative\\path.txt'
    result = string.unix_path_to_wine(rel_path)
    assert result == expected


def test_unix_path_to_wine_path_with_forward_slashes(tmp_path: Path) -> None:
    file_path = tmp_path / 'dir' / 'file.txt'
    file_path.parent.mkdir(exist_ok=True)
    file_path.write_text('test')
    unix_style = str(file_path).replace('\\', '/')
    result = string.unix_path_to_wine(unix_style)
    assert '\\' in result
    assert 'file.txt' in result


@pytest.mark.parametrize(
    ('input_str', 'expected'),
    [
        ('http://example.com', True),
        ('https://example.com', True),
        ('ftp://example.com', True),
        ('not_a_url', False),
        ('just some text', False),
        ('example.com', False),
        ('http:/example.com', False),  # only one slash
        ('file://localhost/etc/fstab', True),
        ('custom_proto123://foo', True),
        ('', False),
    ],
)
def test_is_url_various_cases(input_str: str, expected: bool) -> None:  # noqa: FBT001
    assert string.is_url(input_str) is expected


@pytest.mark.parametrize(
    ('input_str', 'expected'),
    [
        ('ＡＢＣ１２３', 'ABC123'),
        ('！＠＃＄％', '!@#$%'),
        # cspell: disable-next-line  # noqa: ERA001
        ('ａｂｃｄｅｆ', 'abcdef'),
        ('￥１０００', '¥1000'),
        ('Hello　World！', 'Hello World!'),
    ],
)
def test_fullwidth_to_narrow_various_cases(input_str: str, expected: str) -> None:
    assert string.fullwidth_to_narrow(input_str) == expected


@pytest.mark.parametrize(
    ('input_str', 'expected'),
    [
        ('Hello World!', 'hello-world'),
        ('Python_is_awesome', 'python-is-awesome'),
        ('  spaces   and---underscores__ ', 'spaces-and-underscores-'),
        ('Special@#%&* Chars', 'special-chars'),
        ('MiXeD CaSe and 123', 'mixed-case-and-123'),
    ],
)
def test_slugify_various_cases(input_str: str, expected: str) -> None:
    assert string.slugify(input_str) == expected


@pytest.mark.parametrize(
    ('input_str', 'expected'),
    [
        ('XIV', True),  # valid Roman numeral
        # cspell: disable-next-line  # noqa: ERA001
        ('MMXXIV', True),  # valid Roman numeral (2024)
        ('IIII', False),  # invalid (should be IV)
        ('abc', False),  # not a Roman numeral
        ('', False),  # empty string
    ],
)
def test_is_roman_numeral_various_cases(input_str: str, expected: bool) -> None:  # noqa: FBT001
    assert string.is_roman_numeral(input_str) is expected


@pytest.mark.parametrize(
    ('input_sentence', 'expected'),
    [
        ('Hello world.', 'World hello.'),
        ('This is a test!', 'Test a is this!'),
        ('I am here?', 'Here am I?'),
        ('Python is fun', 'Fun is python.'),
        ('Why not try?', 'Try not why?'),
    ],
)
def test_rev_sentence_various_cases(input_sentence: str, expected: str) -> None:
    assert string.rev_sentence(input_sentence) == expected


@pytest.mark.parametrize(
    ('input_sentences', 'expected'),
    [
        (['Hello world.'], ['World hello.']),
        (['This is a test!', 'Another sentence?'], ['Test a is this!', 'Sentence another?']),
        (['I am here?', 'Python is fun'], ['Here am I?', 'Fun is python.']),
        ([''], []),  # empty string yields nothing
        (['   '], []),  # whitespace-only yields nothing
        (['Why not try?', '', 'Hello world.'], ['Try not why?', 'World hello.']),
        (['A single word.'], ['Word single a.']),
        (['i am here.'], ['Here am I.']),  # test 'i' capitalization
        (['Multiple   spaces here!'], ['Here spaces multiple!']),
    ],
)
def test_rev_sentences_various_cases(input_sentences: list[str], expected: list[str]) -> None:
    result = list(string.rev_sentences(input_sentences))
    assert result == expected


def test_sanitize_calls_yt_dlp_sanitize_filename_restricted_true(mocker: MockerFixture) -> None:
    fake_sanitize = mocker.Mock(return_value='Safe-File-Name')
    mocker.patch('deltona.string._get_yt_dlp_sanitize_filename', return_value=fake_sanitize)
    result = string.sanitize('Some Unsafe/File-Name.txt', restricted=True)
    fake_sanitize.assert_called_once_with('Some Unsafe/File-Name.txt', restricted=True)
    # Should be lowercased and dashes normalized
    assert result == 'safe-file-name'


def test_sanitize_calls_yt_dlp_sanitize_filename_restricted_false(mocker: MockerFixture) -> None:
    fake_sanitize = mocker.Mock(return_value='Another_File-Name')
    mocker.patch('deltona.string._get_yt_dlp_sanitize_filename', return_value=fake_sanitize)
    result = string.sanitize('Another*File-Name', restricted=False)
    fake_sanitize.assert_called_once_with('Another*File-Name', restricted=False)
    assert result == 'another-file-name'


def test_sanitize_returns_minimum_dash_if_empty(mocker: MockerFixture) -> None:
    fake_sanitize = mocker.Mock(return_value='_')
    mocker.patch('deltona.string._get_yt_dlp_sanitize_filename', return_value=fake_sanitize)
    result = string.sanitize('', restricted=True)
    assert result == '-'


def test_sanitize_removes_dot_dash_and_normalizes(mocker: MockerFixture) -> None:
    fake_sanitize = mocker.Mock(return_value='file.-name--test')
    mocker.patch('deltona.string._get_yt_dlp_sanitize_filename', return_value=fake_sanitize)
    result = string.sanitize('file.-name--test', restricted=True)
    assert result == 'file-name-test'


def test_sanitize_replaces_pattern(mocker: MockerFixture) -> None:
    fake_sanitize = mocker.Mock(return_value='abc-s-def')
    mocker.patch('deltona.string._get_yt_dlp_sanitize_filename', return_value=fake_sanitize)
    result = string.sanitize('abc-s-def', restricted=True)
    # The ([a-z0-9])\-s\- pattern should be replaced with \1s-
    assert result == 'abcs-def'


def test_add_unidecode_custom_replacement_adds_to_cache(mocker: MockerFixture) -> None:
    fake_cache: dict[int, Sequence[str | None] | None] = {}
    fake_unidecode = mocker.Mock()
    mocker.patch(
        'deltona.string._get_unidecode_cache_and_unidecode',
        return_value=(fake_cache, fake_unidecode),
    )
    mocker.patch('deltona.string.assert_not_none', side_effect=lambda x: x)
    # Add replacement for 'ø'
    find = 'ø'
    replace = 'oe'
    codepoint = ord(find)
    section = codepoint >> 8
    position = codepoint % 256
    # Simulate cache section as None
    fake_cache[section] = None
    string.add_unidecode_custom_replacement(find, replace)
    assert isinstance(fake_cache[section], list)
    assert fake_cache[section][position] == replace  # type: ignore[index]


def test_add_unidecode_custom_replacement_overwrites_existing(mocker: MockerFixture) -> None:
    fake_cache: dict[int, Sequence[str | None] | None] = {}
    fake_unidecode = mocker.Mock()
    mocker.patch(
        'deltona.string._get_unidecode_cache_and_unidecode',
        return_value=(fake_cache, fake_unidecode),
    )
    mocker.patch('deltona.string.assert_not_none', side_effect=lambda x: x)
    find = 'ø'
    replace = 'oe'
    codepoint = ord(find)
    section = codepoint >> 8
    position = codepoint % 256
    # Simulate cache section as a list with a previous value
    fake_cache[section] = [None] * (position + 1)
    fake_cache[section][position] = 'old'  # type: ignore[index]
    string.add_unidecode_custom_replacement(find, replace)
    assert fake_cache[section][position] == replace  # type: ignore[index]


def test_add_unidecode_custom_replacement_handles_existing_tuple(mocker: MockerFixture) -> None:
    fake_cache: dict[int, Sequence[str | None] | None] = {}
    fake_unidecode = mocker.Mock()
    mocker.patch(
        'deltona.string._get_unidecode_cache_and_unidecode',
        return_value=(fake_cache, fake_unidecode),
    )
    mocker.patch('deltona.string.assert_not_none', side_effect=lambda x: x)
    find = 'ø'
    replace = 'oe'
    codepoint = ord(find)
    section = codepoint >> 8
    position = codepoint % 256
    # Simulate cache section as a tuple (immutable)
    fake_cache[section] = tuple([None] * (position + 1))
    string.add_unidecode_custom_replacement(find, replace)
    assert isinstance(fake_cache[section], list)
    assert fake_cache[section][position] == replace  # type: ignore[index]


def test_add_unidecode_custom_replacement_calls_unidecode(mocker: MockerFixture) -> None:
    fake_cache: dict[int, Sequence[str | None] | None] = {}
    fake_unidecode = mocker.Mock()
    mocker.patch(
        'deltona.string._get_unidecode_cache_and_unidecode',
        return_value=(fake_cache, fake_unidecode),
    )
    mocker.patch('deltona.string.assert_not_none', side_effect=lambda x: x)
    find = 'ø'
    replace = 'oe'
    fake_cache[ord(find) >> 8] = None
    string.add_unidecode_custom_replacement(find, replace)
    fake_unidecode.assert_called_once_with(find)


def test_cssq_one_returns_text_content(mocker: MockerFixture) -> None:
    fake_tag1 = mocker.Mock()
    fake_tag1.get_text.return_value = 'First'
    mock_soup = mocker.patch('deltona.string.BeautifulSoup')
    mock_soup.return_value.select_one.return_value = fake_tag1
    result = string.cssq_one('selector', '', text=True)
    assert result == 'First'
    mock_soup.return_value.select_one.assert_called_once_with('selector', flags=0)
    fake_tag1.get_text.assert_called_once_with(strip=True)


def test_cssq_one_returns_none(mocker: MockerFixture) -> None:
    mock_soup = mocker.patch('deltona.string.BeautifulSoup')
    mock_soup.return_value.select_one.return_value = None
    result = string.cssq_one('selector', '')
    assert result is None
    mock_soup.return_value.select_one.assert_called_once_with('selector', flags=0)


def test_cssq_returns_text_content(mocker: MockerFixture) -> None:
    fake_tag1 = mocker.Mock()
    fake_tag1.get_text.return_value = 'First'
    fake_tag2 = mocker.Mock()
    fake_tag2.get_text.return_value = 'Second'
    mocker.patch('deltona.string.BeautifulSoup')
    mock_iselect = mocker.patch('deltona.string.sv.iselect', return_value=[fake_tag1, fake_tag2])
    results = list(string.cssq('selector', '', limit=0, text=True))
    assert results == ['First', 'Second']
    mock_iselect.assert_called_once_with('selector', mocker.ANY, flags=0, limit=0)
    fake_tag1.get_text.assert_called_once_with(strip=True)
    fake_tag2.get_text.assert_called_once_with(strip=True)
